from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from .array_ import fill
from .array_ import fold as fold_1
from .list import FSharpList, cons, head, of_array_with_tail, tail
from .list import empty as empty_1
from .list import fold as fold_2
from .list import is_empty as is_empty_1
from .list import singleton as singleton_1
from .mutable_set import HashSet, HashSet__ctor_Z6150332D, HashSet__get_Comparer
from .option import some
from .option import value as value_3
from .reflection import TypeInfo, bool_type, class_type, list_type, option_type, record_type
from .seq import cache, reduce
from .seq import exists as exists_1
from .seq import fold as fold_3
from .seq import for_all as for_all_1
from .seq import iterate as iterate_1
from .seq import map as map_1
from .string_ import join
from .types import Array, Record, to_string
from .util import (
    IComparer_1,
    IEnumerable_1,
    IEnumerator,
    get_enumerator,
    ignore,
    is_array_like,
    structural_hash,
    to_iterator,
)


_T = TypeVar("_T")

__A = TypeVar("__A")

__B = TypeVar("__B")

_A = TypeVar("_A")

_A_ = TypeVar("_A_")

_U = TypeVar("_U")

_STATE = TypeVar("_STATE")


def _expr231(gen0: TypeInfo) -> TypeInfo:
    return class_type("Set.SetTreeLeaf`1", [gen0], SetTreeLeaf_1)


class SetTreeLeaf_1(Generic[_T]):
    def __init__(self, k: Any | None = None) -> None:
        self.k: _T = k


SetTreeLeaf_1_reflection = _expr231


def SetTreeLeaf_1__ctor_2B595(k: Any | None = None) -> SetTreeLeaf_1[_T]:
    return SetTreeLeaf_1(k)


def SetTreeLeaf_1__get_Key(_: SetTreeLeaf_1[Any]) -> Any:
    return _.k


def _expr232(gen0: TypeInfo) -> TypeInfo:
    return class_type("Set.SetTreeNode`1", [gen0], SetTreeNode_1, SetTreeLeaf_1_reflection(gen0))


class SetTreeNode_1(SetTreeLeaf_1, Generic[_T]):
    def __init__(self, v: Any, left: SetTreeLeaf_1[Any] | None, right: SetTreeLeaf_1[Any] | None, h: int) -> None:
        super().__init__(v)
        self.left: SetTreeLeaf_1[_T] | None = left
        self.right: SetTreeLeaf_1[_T] | None = right
        self.h: int = h or 0


SetTreeNode_1_reflection = _expr232


def SetTreeNode_1__ctor_5F465FC9(
    v: Any, left: SetTreeLeaf_1[Any] | None, right: SetTreeLeaf_1[Any] | None, h: int
) -> SetTreeNode_1[_T]:
    return SetTreeNode_1(v, left, right, h)


def SetTreeNode_1__get_Left(_: SetTreeNode_1[Any]) -> SetTreeLeaf_1[Any] | None:
    return _.left


def SetTreeNode_1__get_Right(_: SetTreeNode_1[Any]) -> SetTreeLeaf_1[Any] | None:
    return _.right


def SetTreeNode_1__get_Height(_: SetTreeNode_1[Any]) -> int:
    return _.h


def SetTreeModule_empty(__unit: None = None) -> SetTreeLeaf_1[Any] | None:
    return None


def SetTreeModule_countAux(t_mut: SetTreeLeaf_1[Any] | None, acc_mut: int) -> int:
    while True:
        (t, acc) = (t_mut, acc_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                t_mut = SetTreeNode_1__get_Left(t2)
                acc_mut = SetTreeModule_countAux(SetTreeNode_1__get_Right(t2), acc + 1)
                continue

            else:
                return acc + 1

        else:
            return acc

        break


def SetTreeModule_count(s: SetTreeLeaf_1[Any] | None = None) -> int:
    return SetTreeModule_countAux(s, 0)


def SetTreeModule_mk(
    l: SetTreeLeaf_1[Any] | None, k: Any, r: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    hl: int
    t: SetTreeLeaf_1[_T] | None = l
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        hl = SetTreeNode_1__get_Height(t2) if isinstance(t2, SetTreeNode_1) else 1

    else:
        hl = 0

    hr: int
    t_1: SetTreeLeaf_1[_T] | None = r
    if t_1 is not None:
        t2_1: SetTreeLeaf_1[_T] = t_1
        hr = SetTreeNode_1__get_Height(t2_1) if isinstance(t2_1, SetTreeNode_1) else 1

    else:
        hr = 0

    m: int = (hr if (hl < hr) else hl) or 0
    if m == 0:
        return SetTreeLeaf_1__ctor_2B595(k)

    else:
        return SetTreeNode_1__ctor_5F465FC9(k, l, r, m + 1)


def SetTreeModule_rebalance(
    t1: SetTreeLeaf_1[Any] | None, v: Any, t2: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    t1h: int
    t: SetTreeLeaf_1[_T] | None = t1
    if t is not None:
        t2_1: SetTreeLeaf_1[_T] = t
        t1h = SetTreeNode_1__get_Height(t2_1) if isinstance(t2_1, SetTreeNode_1) else 1

    else:
        t1h = 0

    t2h: int
    t_1: SetTreeLeaf_1[_T] | None = t2
    if t_1 is not None:
        t2_2: SetTreeLeaf_1[_T] = t_1
        t2h = SetTreeNode_1__get_Height(t2_2) if isinstance(t2_2, SetTreeNode_1) else 1

    else:
        t2h = 0

    if t2h > (t1h + 2):
        match_value: SetTreeLeaf_1[_T] = value_3(t2)
        if isinstance(match_value, SetTreeNode_1):

            def _arrow233(__unit: None = None, t1: Any = t1, v: Any = v, t2: Any = t2) -> int:
                t_2: SetTreeLeaf_1[_T] | None = SetTreeNode_1__get_Left(match_value)
                if t_2 is not None:
                    t2_3: SetTreeLeaf_1[_T] = t_2
                    return SetTreeNode_1__get_Height(t2_3) if isinstance(t2_3, SetTreeNode_1) else 1

                else:
                    return 0

            if _arrow233() > (t1h + 1):
                match_value_1: SetTreeLeaf_1[_T] = value_3(SetTreeNode_1__get_Left(match_value))
                if isinstance(match_value_1, SetTreeNode_1):
                    return SetTreeModule_mk(
                        SetTreeModule_mk(t1, v, SetTreeNode_1__get_Left(match_value_1)),
                        SetTreeLeaf_1__get_Key(match_value_1),
                        SetTreeModule_mk(
                            SetTreeNode_1__get_Right(match_value_1),
                            SetTreeLeaf_1__get_Key(match_value),
                            SetTreeNode_1__get_Right(match_value),
                        ),
                    )

                else:
                    raise Exception("internal error: Set.rebalance")

            else:
                return SetTreeModule_mk(
                    SetTreeModule_mk(t1, v, SetTreeNode_1__get_Left(match_value)),
                    SetTreeLeaf_1__get_Key(match_value),
                    SetTreeNode_1__get_Right(match_value),
                )

        else:
            raise Exception("internal error: Set.rebalance")

    elif t1h > (t2h + 2):
        match_value_2: SetTreeLeaf_1[_T] = value_3(t1)
        if isinstance(match_value_2, SetTreeNode_1):

            def _arrow234(__unit: None = None, t1: Any = t1, v: Any = v, t2: Any = t2) -> int:
                t_3: SetTreeLeaf_1[_T] | None = SetTreeNode_1__get_Right(match_value_2)
                if t_3 is not None:
                    t2_4: SetTreeLeaf_1[_T] = t_3
                    return SetTreeNode_1__get_Height(t2_4) if isinstance(t2_4, SetTreeNode_1) else 1

                else:
                    return 0

            if _arrow234() > (t2h + 1):
                match_value_3: SetTreeLeaf_1[_T] = value_3(SetTreeNode_1__get_Right(match_value_2))
                if isinstance(match_value_3, SetTreeNode_1):
                    return SetTreeModule_mk(
                        SetTreeModule_mk(
                            SetTreeNode_1__get_Left(match_value_2),
                            SetTreeLeaf_1__get_Key(match_value_2),
                            SetTreeNode_1__get_Left(match_value_3),
                        ),
                        SetTreeLeaf_1__get_Key(match_value_3),
                        SetTreeModule_mk(SetTreeNode_1__get_Right(match_value_3), v, t2),
                    )

                else:
                    raise Exception("internal error: Set.rebalance")

            else:
                return SetTreeModule_mk(
                    SetTreeNode_1__get_Left(match_value_2),
                    SetTreeLeaf_1__get_Key(match_value_2),
                    SetTreeModule_mk(SetTreeNode_1__get_Right(match_value_2), v, t2),
                )

        else:
            raise Exception("internal error: Set.rebalance")

    else:
        return SetTreeModule_mk(t1, v, t2)


def SetTreeModule_add(
    comparer: IComparer_1[Any], k: Any, t: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        c: int = comparer.Compare(k, SetTreeLeaf_1__get_Key(t2)) or 0
        if isinstance(t2, SetTreeNode_1):
            if c < 0:
                return SetTreeModule_rebalance(
                    SetTreeModule_add(comparer, k, SetTreeNode_1__get_Left(t2)),
                    SetTreeLeaf_1__get_Key(t2),
                    SetTreeNode_1__get_Right(t2),
                )

            elif c == 0:
                return t

            else:
                return SetTreeModule_rebalance(
                    SetTreeNode_1__get_Left(t2),
                    SetTreeLeaf_1__get_Key(t2),
                    SetTreeModule_add(comparer, k, SetTreeNode_1__get_Right(t2)),
                )

        else:
            c_1: int = comparer.Compare(k, SetTreeLeaf_1__get_Key(t2)) or 0
            if c_1 < 0:
                return SetTreeNode_1__ctor_5F465FC9(k, SetTreeModule_empty(), t, 2)

            elif c_1 == 0:
                return t

            else:
                return SetTreeNode_1__ctor_5F465FC9(k, t, SetTreeModule_empty(), 2)

    else:
        return SetTreeLeaf_1__ctor_2B595(k)


def SetTreeModule_balance(
    comparer: IComparer_1[Any], t1: SetTreeLeaf_1[Any] | None, k: Any, t2: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    if t1 is not None:
        t1_0027: SetTreeLeaf_1[_T] = t1
        if t2 is not None:
            t2_0027: SetTreeLeaf_1[_T] = t2
            if isinstance(t1_0027, SetTreeNode_1):
                if isinstance(t2_0027, SetTreeNode_1):
                    if (SetTreeNode_1__get_Height(t1_0027) + 2) < SetTreeNode_1__get_Height(t2_0027):
                        return SetTreeModule_rebalance(
                            SetTreeModule_balance(comparer, t1, k, SetTreeNode_1__get_Left(t2_0027)),
                            SetTreeLeaf_1__get_Key(t2_0027),
                            SetTreeNode_1__get_Right(t2_0027),
                        )

                    elif (SetTreeNode_1__get_Height(t2_0027) + 2) < SetTreeNode_1__get_Height(t1_0027):
                        return SetTreeModule_rebalance(
                            SetTreeNode_1__get_Left(t1_0027),
                            SetTreeLeaf_1__get_Key(t1_0027),
                            SetTreeModule_balance(comparer, SetTreeNode_1__get_Right(t1_0027), k, t2),
                        )

                    else:
                        return SetTreeModule_mk(t1, k, t2)

                else:
                    return SetTreeModule_add(
                        comparer, k, SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2_0027), t1)
                    )

            else:
                return SetTreeModule_add(comparer, k, SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t1_0027), t2))

        else:
            return SetTreeModule_add(comparer, k, t1)

    else:
        return SetTreeModule_add(comparer, k, t2)


def SetTreeModule_split(
    comparer: IComparer_1[Any], pivot: Any, t: SetTreeLeaf_1[Any] | None = None
) -> tuple[SetTreeLeaf_1[_T] | None, bool, SetTreeLeaf_1[_T] | None]:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        if isinstance(t2, SetTreeNode_1):
            c: int = comparer.Compare(pivot, SetTreeLeaf_1__get_Key(t2)) or 0
            if c < 0:
                pattern_input: tuple[SetTreeLeaf_1[_T] | None, bool, SetTreeLeaf_1[_T] | None] = SetTreeModule_split(
                    comparer, pivot, SetTreeNode_1__get_Left(t2)
                )
                return (
                    pattern_input[0],
                    pattern_input[1],
                    SetTreeModule_balance(
                        comparer, pattern_input[2], SetTreeLeaf_1__get_Key(t2), SetTreeNode_1__get_Right(t2)
                    ),
                )

            elif c == 0:
                return (SetTreeNode_1__get_Left(t2), True, SetTreeNode_1__get_Right(t2))

            else:
                pattern_input_1: tuple[SetTreeLeaf_1[_T] | None, bool, SetTreeLeaf_1[_T] | None] = SetTreeModule_split(
                    comparer, pivot, SetTreeNode_1__get_Right(t2)
                )
                return (
                    SetTreeModule_balance(
                        comparer, SetTreeNode_1__get_Left(t2), SetTreeLeaf_1__get_Key(t2), pattern_input_1[0]
                    ),
                    pattern_input_1[1],
                    pattern_input_1[2],
                )

        else:
            c_1: int = comparer.Compare(SetTreeLeaf_1__get_Key(t2), pivot) or 0
            if c_1 < 0:
                return (t, False, SetTreeModule_empty())

            elif c_1 == 0:
                return (SetTreeModule_empty(), True, SetTreeModule_empty())

            else:
                return (SetTreeModule_empty(), False, t)

    else:
        return (SetTreeModule_empty(), False, SetTreeModule_empty())


def SetTreeModule_spliceOutSuccessor(t: SetTreeLeaf_1[Any] | None = None) -> tuple[_T, SetTreeLeaf_1[_T] | None]:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        if isinstance(t2, SetTreeNode_1):
            if SetTreeNode_1__get_Left(t2) is None:
                return (SetTreeLeaf_1__get_Key(t2), SetTreeNode_1__get_Right(t2))

            else:
                pattern_input: tuple[_T, SetTreeLeaf_1[_T] | None] = SetTreeModule_spliceOutSuccessor(
                    SetTreeNode_1__get_Left(t2)
                )
                return (
                    pattern_input[0],
                    SetTreeModule_mk(pattern_input[1], SetTreeLeaf_1__get_Key(t2), SetTreeNode_1__get_Right(t2)),
                )

        else:
            return (SetTreeLeaf_1__get_Key(t2), SetTreeModule_empty())

    else:
        raise Exception("internal error: Set.spliceOutSuccessor")


def SetTreeModule_remove(
    comparer: IComparer_1[Any], k: Any, t: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        c: int = comparer.Compare(k, SetTreeLeaf_1__get_Key(t2)) or 0
        if isinstance(t2, SetTreeNode_1):
            if c < 0:
                return SetTreeModule_rebalance(
                    SetTreeModule_remove(comparer, k, SetTreeNode_1__get_Left(t2)),
                    SetTreeLeaf_1__get_Key(t2),
                    SetTreeNode_1__get_Right(t2),
                )

            elif c == 0:
                if SetTreeNode_1__get_Left(t2) is None:
                    return SetTreeNode_1__get_Right(t2)

                elif SetTreeNode_1__get_Right(t2) is None:
                    return SetTreeNode_1__get_Left(t2)

                else:
                    pattern_input: tuple[_T, SetTreeLeaf_1[_T] | None] = SetTreeModule_spliceOutSuccessor(
                        SetTreeNode_1__get_Right(t2)
                    )
                    return SetTreeModule_mk(SetTreeNode_1__get_Left(t2), pattern_input[0], pattern_input[1])

            else:
                return SetTreeModule_rebalance(
                    SetTreeNode_1__get_Left(t2),
                    SetTreeLeaf_1__get_Key(t2),
                    SetTreeModule_remove(comparer, k, SetTreeNode_1__get_Right(t2)),
                )

        elif c == 0:
            return SetTreeModule_empty()

        else:
            return t

    else:
        return t


def SetTreeModule_mem(comparer_mut: IComparer_1[Any], k_mut: Any, t_mut: SetTreeLeaf_1[Any] | None) -> bool:
    while True:
        (comparer, k, t) = (comparer_mut, k_mut, t_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            c: int = comparer.Compare(k, SetTreeLeaf_1__get_Key(t2)) or 0
            if isinstance(t2, SetTreeNode_1):
                if c < 0:
                    comparer_mut = comparer
                    k_mut = k
                    t_mut = SetTreeNode_1__get_Left(t2)
                    continue

                elif c == 0:
                    return True

                else:
                    comparer_mut = comparer
                    k_mut = k
                    t_mut = SetTreeNode_1__get_Right(t2)
                    continue

            else:
                return c == 0

        else:
            return False

        break


def SetTreeModule_iter(f_mut: Callable[[_T], None], t_mut: SetTreeLeaf_1[Any] | None) -> None:
    while True:
        (f, t) = (f_mut, t_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                SetTreeModule_iter(f, SetTreeNode_1__get_Left(t2))
                f(SetTreeLeaf_1__get_Key(t2))
                f_mut = f
                t_mut = SetTreeNode_1__get_Right(t2)
                continue

            else:
                f(SetTreeLeaf_1__get_Key(t2))

        break


def SetTreeModule_foldBackOpt(f_mut: Any, t_mut: SetTreeLeaf_1[Any] | None, x_mut: Any) -> Any:
    while True:
        (f, t, x) = (f_mut, t_mut, x_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                f_mut = f
                t_mut = SetTreeNode_1__get_Left(t2)
                x_mut = f(SetTreeLeaf_1__get_Key(t2), SetTreeModule_foldBackOpt(f, SetTreeNode_1__get_Right(t2), x))
                continue

            else:
                return f(SetTreeLeaf_1__get_Key(t2), x)

        else:
            return x

        break


def SetTreeModule_foldBack(f: Callable[[__A, __B], __B], m: SetTreeLeaf_1[Any] | None, x: Any) -> Any:
    return SetTreeModule_foldBackOpt(f, m, x)


def SetTreeModule_foldOpt(f_mut: Any, x_mut: Any, t_mut: SetTreeLeaf_1[Any] | None) -> Any:
    while True:
        (f, x, t) = (f_mut, x_mut, t_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                f_mut = f
                x_mut = f(SetTreeModule_foldOpt(f, x, SetTreeNode_1__get_Left(t2)), SetTreeLeaf_1__get_Key(t2))
                t_mut = SetTreeNode_1__get_Right(t2)
                continue

            else:
                return f(x, SetTreeLeaf_1__get_Key(t2))

        else:
            return x

        break


def SetTreeModule_fold(f: Callable[[__A, __B], __A], x: Any, m: SetTreeLeaf_1[Any] | None = None) -> Any:
    return SetTreeModule_foldOpt(f, x, m)


def SetTreeModule_forall(f_mut: Callable[[_T], bool], t_mut: SetTreeLeaf_1[Any] | None) -> bool:
    while True:
        (f, t) = (f_mut, t_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                if SetTreeModule_forall(f, SetTreeNode_1__get_Left(t2)) if f(SetTreeLeaf_1__get_Key(t2)) else False:
                    f_mut = f
                    t_mut = SetTreeNode_1__get_Right(t2)
                    continue

                else:
                    return False

            else:
                return f(SetTreeLeaf_1__get_Key(t2))

        else:
            return True

        break


def SetTreeModule_exists(f_mut: Callable[[_T], bool], t_mut: SetTreeLeaf_1[Any] | None) -> bool:
    while True:
        (f, t) = (f_mut, t_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                if True if f(SetTreeLeaf_1__get_Key(t2)) else SetTreeModule_exists(f, SetTreeNode_1__get_Left(t2)):
                    return True

                else:
                    f_mut = f
                    t_mut = SetTreeNode_1__get_Right(t2)
                    continue

            else:
                return f(SetTreeLeaf_1__get_Key(t2))

        else:
            return False

        break


def SetTreeModule_subset(
    comparer: IComparer_1[Any], a: SetTreeLeaf_1[Any] | None = None, b: SetTreeLeaf_1[Any] | None = None
) -> bool:
    def _arrow235(x: __A | None = None, comparer: Any = comparer, a: Any = a, b: Any = b) -> bool:
        return SetTreeModule_mem(comparer, x, b)

    return SetTreeModule_forall(_arrow235, a)


def SetTreeModule_properSubset(
    comparer: IComparer_1[Any], a: SetTreeLeaf_1[Any] | None = None, b: SetTreeLeaf_1[Any] | None = None
) -> bool:
    def _arrow236(x: __A | None = None, comparer: Any = comparer, a: Any = a, b: Any = b) -> bool:
        return SetTreeModule_mem(comparer, x, b)

    if SetTreeModule_forall(_arrow236, a):

        def _arrow237(x_1: __A | None = None, comparer: Any = comparer, a: Any = a, b: Any = b) -> bool:
            return not SetTreeModule_mem(comparer, x_1, a)

        return SetTreeModule_exists(_arrow237, b)

    else:
        return False


def SetTreeModule_filterAux(
    comparer_mut: IComparer_1[Any],
    f_mut: Callable[[_T], bool],
    t_mut: SetTreeLeaf_1[Any] | None,
    acc_mut: SetTreeLeaf_1[Any] | None,
) -> SetTreeLeaf_1[Any] | None:
    while True:
        (comparer, f, t, acc) = (comparer_mut, f_mut, t_mut, acc_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                acc_1: SetTreeLeaf_1[_T] | None = (
                    SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2), acc)
                    if f(SetTreeLeaf_1__get_Key(t2))
                    else acc
                )
                comparer_mut = comparer
                f_mut = f
                t_mut = SetTreeNode_1__get_Left(t2)
                acc_mut = SetTreeModule_filterAux(comparer, f, SetTreeNode_1__get_Right(t2), acc_1)
                continue

            elif f(SetTreeLeaf_1__get_Key(t2)):
                return SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2), acc)

            else:
                return acc

        else:
            return acc

        break


def SetTreeModule_filter(
    comparer: IComparer_1[Any], f: Callable[[__A], bool], s: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    return SetTreeModule_filterAux(comparer, f, s, SetTreeModule_empty())


def SetTreeModule_diffAux(
    comparer_mut: IComparer_1[Any], t_mut: SetTreeLeaf_1[Any] | None, acc_mut: SetTreeLeaf_1[Any] | None
) -> SetTreeLeaf_1[Any] | None:
    while True:
        (comparer, t, acc) = (comparer_mut, t_mut, acc_mut)
        if acc is None:
            return acc

        elif t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                comparer_mut = comparer
                t_mut = SetTreeNode_1__get_Left(t2)
                acc_mut = SetTreeModule_diffAux(
                    comparer,
                    SetTreeNode_1__get_Right(t2),
                    SetTreeModule_remove(comparer, SetTreeLeaf_1__get_Key(t2), acc),
                )
                continue

            else:
                return SetTreeModule_remove(comparer, SetTreeLeaf_1__get_Key(t2), acc)

        else:
            return acc

        break


def SetTreeModule_diff(
    comparer: IComparer_1[Any], a: SetTreeLeaf_1[Any] | None = None, b: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    return SetTreeModule_diffAux(comparer, b, a)


def SetTreeModule_union(
    comparer: IComparer_1[Any], t1: SetTreeLeaf_1[Any] | None = None, t2: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    if t1 is not None:
        t1_0027: SetTreeLeaf_1[_T] = t1
        if t2 is not None:
            t2_0027: SetTreeLeaf_1[_T] = t2
            if isinstance(t1_0027, SetTreeNode_1):
                if isinstance(t2_0027, SetTreeNode_1):
                    if SetTreeNode_1__get_Height(t1_0027) > SetTreeNode_1__get_Height(t2_0027):
                        pattern_input: tuple[SetTreeLeaf_1[_T] | None, bool, SetTreeLeaf_1[_T] | None] = (
                            SetTreeModule_split(comparer, SetTreeLeaf_1__get_Key(t1_0027), t2)
                        )
                        return SetTreeModule_balance(
                            comparer,
                            SetTreeModule_union(comparer, SetTreeNode_1__get_Left(t1_0027), pattern_input[0]),
                            SetTreeLeaf_1__get_Key(t1_0027),
                            SetTreeModule_union(comparer, SetTreeNode_1__get_Right(t1_0027), pattern_input[2]),
                        )

                    else:
                        pattern_input_1: tuple[SetTreeLeaf_1[_T] | None, bool, SetTreeLeaf_1[_T] | None] = (
                            SetTreeModule_split(comparer, SetTreeLeaf_1__get_Key(t2_0027), t1)
                        )
                        return SetTreeModule_balance(
                            comparer,
                            SetTreeModule_union(comparer, SetTreeNode_1__get_Left(t2_0027), pattern_input_1[0]),
                            SetTreeLeaf_1__get_Key(t2_0027),
                            SetTreeModule_union(comparer, SetTreeNode_1__get_Right(t2_0027), pattern_input_1[2]),
                        )

                else:
                    return SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2_0027), t1)

            else:
                return SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t1_0027), t2)

        else:
            return t1

    else:
        return t2


def SetTreeModule_intersectionAux(
    comparer_mut: IComparer_1[Any],
    b_mut: SetTreeLeaf_1[Any] | None,
    t_mut: SetTreeLeaf_1[Any] | None,
    acc_mut: SetTreeLeaf_1[Any] | None,
) -> SetTreeLeaf_1[Any] | None:
    while True:
        (comparer, b, t, acc) = (comparer_mut, b_mut, t_mut, acc_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                acc_1: SetTreeLeaf_1[_T] | None = SetTreeModule_intersectionAux(
                    comparer, b, SetTreeNode_1__get_Right(t2), acc
                )
                acc_2: SetTreeLeaf_1[_T] | None = (
                    SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2), acc_1)
                    if SetTreeModule_mem(comparer, SetTreeLeaf_1__get_Key(t2), b)
                    else acc_1
                )
                comparer_mut = comparer
                b_mut = b
                t_mut = SetTreeNode_1__get_Left(t2)
                acc_mut = acc_2
                continue

            elif SetTreeModule_mem(comparer, SetTreeLeaf_1__get_Key(t2), b):
                return SetTreeModule_add(comparer, SetTreeLeaf_1__get_Key(t2), acc)

            else:
                return acc

        else:
            return acc

        break


def SetTreeModule_intersection(
    comparer: IComparer_1[Any], a: SetTreeLeaf_1[Any] | None = None, b: SetTreeLeaf_1[Any] | None = None
) -> SetTreeLeaf_1[Any] | None:
    return SetTreeModule_intersectionAux(comparer, b, a, SetTreeModule_empty())


def SetTreeModule_partition1(
    comparer: IComparer_1[Any],
    f: Callable[[__A], bool],
    k: Any,
    acc1: SetTreeLeaf_1[Any] | None = None,
    acc2: SetTreeLeaf_1[Any] | None = None,
) -> tuple[SetTreeLeaf_1[__A] | None, SetTreeLeaf_1[__A] | None]:
    if f(k):
        return (SetTreeModule_add(comparer, k, acc1), acc2)

    else:
        return (acc1, SetTreeModule_add(comparer, k, acc2))


def SetTreeModule_partitionAux(
    comparer_mut: IComparer_1[Any],
    f_mut: Callable[[_T], bool],
    t_mut: SetTreeLeaf_1[Any] | None,
    acc__mut: SetTreeLeaf_1[Any] | None,
    acc__1_mut: SetTreeLeaf_1[Any] | None,
) -> tuple[SetTreeLeaf_1[_T] | None, SetTreeLeaf_1[_T] | None]:
    while True:
        (comparer, f, t, acc_, acc__1) = (comparer_mut, f_mut, t_mut, acc__mut, acc__1_mut)
        acc: tuple[SetTreeLeaf_1[_T] | None, SetTreeLeaf_1[_T] | None] = (acc_, acc__1)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                acc_1: tuple[SetTreeLeaf_1[_T] | None, SetTreeLeaf_1[_T] | None] = SetTreeModule_partitionAux(
                    comparer, f, SetTreeNode_1__get_Right(t2), acc[0], acc[1]
                )
                acc_4: tuple[SetTreeLeaf_1[_T] | None, SetTreeLeaf_1[_T] | None] = SetTreeModule_partition1(
                    comparer, f, SetTreeLeaf_1__get_Key(t2), acc_1[0], acc_1[1]
                )
                comparer_mut = comparer
                f_mut = f
                t_mut = SetTreeNode_1__get_Left(t2)
                acc__mut = acc_4[0]
                acc__1_mut = acc_4[1]
                continue

            else:
                return SetTreeModule_partition1(comparer, f, SetTreeLeaf_1__get_Key(t2), acc[0], acc[1])

        else:
            return acc

        break


def SetTreeModule_partition(
    comparer: IComparer_1[Any], f: Callable[[__A], bool], s: SetTreeLeaf_1[Any] | None = None
) -> tuple[SetTreeLeaf_1[__A] | None, SetTreeLeaf_1[__A] | None]:
    return SetTreeModule_partitionAux(comparer, f, s, SetTreeModule_empty(), SetTreeModule_empty())


def SetTreeModule_minimumElementAux(t_mut: SetTreeLeaf_1[Any] | None, n_mut: Any) -> Any:
    while True:
        (t, n) = (t_mut, n_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                t_mut = SetTreeNode_1__get_Left(t2)
                n_mut = SetTreeLeaf_1__get_Key(t2)
                continue

            else:
                return SetTreeLeaf_1__get_Key(t2)

        else:
            return n

        break


def SetTreeModule_minimumElementOpt(t: SetTreeLeaf_1[Any] | None = None) -> Any | None:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        if isinstance(t2, SetTreeNode_1):
            return some(SetTreeModule_minimumElementAux(SetTreeNode_1__get_Left(t2), SetTreeLeaf_1__get_Key(t2)))

        else:
            return some(SetTreeLeaf_1__get_Key(t2))

    else:
        return None


def SetTreeModule_maximumElementAux(t_mut: SetTreeLeaf_1[Any] | None, n_mut: Any) -> Any:
    while True:
        (t, n) = (t_mut, n_mut)
        if t is not None:
            t2: SetTreeLeaf_1[_T] = t
            if isinstance(t2, SetTreeNode_1):
                t_mut = SetTreeNode_1__get_Right(t2)
                n_mut = SetTreeLeaf_1__get_Key(t2)
                continue

            else:
                return SetTreeLeaf_1__get_Key(t2)

        else:
            return n

        break


def SetTreeModule_maximumElementOpt(t: SetTreeLeaf_1[Any] | None = None) -> Any | None:
    if t is not None:
        t2: SetTreeLeaf_1[_T] = t
        if isinstance(t2, SetTreeNode_1):
            return some(SetTreeModule_maximumElementAux(SetTreeNode_1__get_Right(t2), SetTreeLeaf_1__get_Key(t2)))

        else:
            return some(SetTreeLeaf_1__get_Key(t2))

    else:
        return None


def SetTreeModule_minimumElement(s: SetTreeLeaf_1[Any] | None = None) -> Any:
    match_value: __A | None = SetTreeModule_minimumElementOpt(s)
    if match_value is None:
        raise Exception("Set contains no elements")

    else:
        return value_3(match_value)


def SetTreeModule_maximumElement(s: SetTreeLeaf_1[Any] | None = None) -> Any:
    match_value: __A | None = SetTreeModule_maximumElementOpt(s)
    if match_value is None:
        raise Exception("Set contains no elements")

    else:
        return value_3(match_value)


def _expr238(gen0: TypeInfo) -> TypeInfo:
    return record_type(
        "Set.SetTreeModule.SetIterator`1",
        [gen0],
        SetTreeModule_SetIterator_1,
        lambda: [("stack", list_type(option_type(SetTreeLeaf_1_reflection(gen0)))), ("started", bool_type)],
    )


@dataclass(eq=False, repr=False, slots=True)
class SetTreeModule_SetIterator_1(Record, Generic[_T]):
    stack: FSharpList[SetTreeLeaf_1[_T] | None]
    started: bool


SetTreeModule_SetIterator_1_reflection = _expr238


def SetTreeModule_collapseLHS(
    stack_mut: FSharpList[SetTreeLeaf_1[Any] | None],
) -> FSharpList[SetTreeLeaf_1[Any] | None]:
    while True:
        (stack,) = (stack_mut,)
        if not is_empty_1(stack):
            x: SetTreeLeaf_1[_T] | None = head(stack)
            rest: FSharpList[SetTreeLeaf_1[_T] | None] = tail(stack)
            if x is not None:
                x2: SetTreeLeaf_1[_T] = x
                if isinstance(x2, SetTreeNode_1):
                    stack_mut = of_array_with_tail(
                        [
                            SetTreeNode_1__get_Left(x2),
                            SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2)),
                            SetTreeNode_1__get_Right(x2),
                        ],
                        rest,
                    )
                    continue

                else:
                    return stack

            else:
                stack_mut = rest
                continue

        else:
            return empty_1()

        break


def SetTreeModule_mkIterator(s: SetTreeLeaf_1[Any] | None = None) -> SetTreeModule_SetIterator_1[Any]:
    return SetTreeModule_SetIterator_1(SetTreeModule_collapseLHS(singleton_1(s)), False)


def SetTreeModule_notStarted(__unit: None = None) -> Any:
    raise Exception("Enumeration not started")


def SetTreeModule_alreadyFinished(__unit: None = None) -> Any:
    raise Exception("Enumeration already started")


def SetTreeModule_current(i: SetTreeModule_SetIterator_1[Any]) -> Any:
    if i.started:
        match_value: FSharpList[SetTreeLeaf_1[__A] | None] = i.stack
        if is_empty_1(match_value):
            return SetTreeModule_alreadyFinished()

        elif head(match_value) is not None:
            t: SetTreeLeaf_1[__A] = head(match_value)
            return SetTreeLeaf_1__get_Key(t)

        else:
            raise Exception("Please report error: Set iterator, unexpected stack for current")

    else:
        return SetTreeModule_notStarted()


def SetTreeModule_moveNext(i: SetTreeModule_SetIterator_1[Any]) -> bool:
    if i.started:
        match_value: FSharpList[SetTreeLeaf_1[_T] | None] = i.stack
        if not is_empty_1(match_value):
            if head(match_value) is not None:
                t: SetTreeLeaf_1[_T] = head(match_value)
                if isinstance(t, SetTreeNode_1):
                    raise Exception("Please report error: Set iterator, unexpected stack for moveNext")

                else:
                    i.stack = SetTreeModule_collapseLHS(tail(match_value))
                    return not is_empty_1(i.stack)

            else:
                raise Exception("Please report error: Set iterator, unexpected stack for moveNext")

        else:
            return False

    else:
        i.started = True
        return not is_empty_1(i.stack)


def SetTreeModule_mkIEnumerator(s: SetTreeLeaf_1[Any] | None = None) -> IEnumerator[Any]:
    i: SetTreeModule_SetIterator_1[_A] = SetTreeModule_mkIterator(s)

    class ObjectExpr239(IEnumerator[_A_]):
        def System_Collections_Generic_IEnumerator_1_get_Current(self, __unit: None = None, s: Any = s) -> _A_:
            return SetTreeModule_current(i)

        def System_Collections_IEnumerator_get_Current(self, __unit: None = None, s: Any = s) -> Any:
            return SetTreeModule_current(i)

        def System_Collections_IEnumerator_MoveNext(self, __unit: None = None, s: Any = s) -> bool:
            return SetTreeModule_moveNext(i)

        def System_Collections_IEnumerator_Reset(self, __unit: None = None, s: Any = s) -> None:
            nonlocal i
            i = SetTreeModule_mkIterator(s)

        def Dispose(self, __unit: None = None, s: Any = s) -> None:
            pass

    return ObjectExpr239()


def SetTreeModule_compareStacks(
    comparer_mut: IComparer_1[Any],
    l1_mut: FSharpList[SetTreeLeaf_1[Any] | None],
    l2_mut: FSharpList[SetTreeLeaf_1[Any] | None],
) -> int:
    while True:
        (comparer, l1, l2) = (comparer_mut, l1_mut, l2_mut)
        if not is_empty_1(l1):
            if not is_empty_1(l2):
                if head(l2) is not None:
                    if head(l1) is not None:
                        x1_3: SetTreeLeaf_1[_T] = head(l1)
                        x2_3: SetTreeLeaf_1[_T] = head(l2)
                        if isinstance(x1_3, SetTreeNode_1):
                            if SetTreeNode_1__get_Left(x1_3) is None:
                                if isinstance(x2_3, SetTreeNode_1):
                                    if SetTreeNode_1__get_Left(x2_3) is None:
                                        c: int = (
                                            comparer.Compare(SetTreeLeaf_1__get_Key(x1_3), SetTreeLeaf_1__get_Key(x2_3))
                                            or 0
                                        )
                                        if c != 0:
                                            return c

                                        else:
                                            comparer_mut = comparer
                                            l1_mut = cons(SetTreeNode_1__get_Right(x1_3), tail(l1))
                                            l2_mut = cons(SetTreeNode_1__get_Right(x2_3), tail(l2))
                                            continue

                                    else:
                                        (pattern_matching_result, t1_6, x1_4, t2_6, x2_4) = (
                                            None,
                                            None,
                                            None,
                                            None,
                                            None,
                                        )
                                        if not is_empty_1(l1):
                                            if head(l1) is not None:
                                                pattern_matching_result = 0
                                                t1_6 = tail(l1)
                                                x1_4 = head(l1)

                                            elif not is_empty_1(l2):
                                                if head(l2) is not None:
                                                    pattern_matching_result = 1
                                                    t2_6 = tail(l2)
                                                    x2_4 = head(l2)

                                                else:
                                                    pattern_matching_result = 2

                                            else:
                                                pattern_matching_result = 2

                                        elif not is_empty_1(l2):
                                            if head(l2) is not None:
                                                pattern_matching_result = 1
                                                t2_6 = tail(l2)
                                                x2_4 = head(l2)

                                            else:
                                                pattern_matching_result = 2

                                        else:
                                            pattern_matching_result = 2

                                        if pattern_matching_result == 0:
                                            if isinstance(x1_4, SetTreeNode_1):
                                                comparer_mut = comparer
                                                l1_mut = of_array_with_tail(
                                                    [
                                                        SetTreeNode_1__get_Left(x1_4),
                                                        SetTreeNode_1__ctor_5F465FC9(
                                                            SetTreeLeaf_1__get_Key(x1_4),
                                                            SetTreeModule_empty(),
                                                            SetTreeNode_1__get_Right(x1_4),
                                                            0,
                                                        ),
                                                    ],
                                                    t1_6,
                                                )
                                                l2_mut = l2
                                                continue

                                            else:
                                                comparer_mut = comparer
                                                l1_mut = of_array_with_tail(
                                                    [
                                                        SetTreeModule_empty(),
                                                        SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x1_4)),
                                                    ],
                                                    t1_6,
                                                )
                                                l2_mut = l2
                                                continue

                                        elif pattern_matching_result == 1:
                                            if isinstance(x2_4, SetTreeNode_1):
                                                comparer_mut = comparer
                                                l1_mut = l1
                                                l2_mut = of_array_with_tail(
                                                    [
                                                        SetTreeNode_1__get_Left(x2_4),
                                                        SetTreeNode_1__ctor_5F465FC9(
                                                            SetTreeLeaf_1__get_Key(x2_4),
                                                            SetTreeModule_empty(),
                                                            SetTreeNode_1__get_Right(x2_4),
                                                            0,
                                                        ),
                                                    ],
                                                    t2_6,
                                                )
                                                continue

                                            else:
                                                comparer_mut = comparer
                                                l1_mut = l1
                                                l2_mut = of_array_with_tail(
                                                    [
                                                        SetTreeModule_empty(),
                                                        SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2_4)),
                                                    ],
                                                    t2_6,
                                                )
                                                continue

                                        elif pattern_matching_result == 2:
                                            raise Exception("unexpected state in SetTree.compareStacks")

                                else:
                                    c_1: int = (
                                        comparer.Compare(SetTreeLeaf_1__get_Key(x1_3), SetTreeLeaf_1__get_Key(x2_3))
                                        or 0
                                    )
                                    if c_1 != 0:
                                        return c_1

                                    else:
                                        comparer_mut = comparer
                                        l1_mut = cons(SetTreeNode_1__get_Right(x1_3), tail(l1))
                                        l2_mut = cons(SetTreeModule_empty(), tail(l2))
                                        continue

                            else:
                                (pattern_matching_result_1, t1_7, x1_5, t2_7, x2_5) = (None, None, None, None, None)
                                if not is_empty_1(l1):
                                    if head(l1) is not None:
                                        pattern_matching_result_1 = 0
                                        t1_7 = tail(l1)
                                        x1_5 = head(l1)

                                    elif not is_empty_1(l2):
                                        if head(l2) is not None:
                                            pattern_matching_result_1 = 1
                                            t2_7 = tail(l2)
                                            x2_5 = head(l2)

                                        else:
                                            pattern_matching_result_1 = 2

                                    else:
                                        pattern_matching_result_1 = 2

                                elif not is_empty_1(l2):
                                    if head(l2) is not None:
                                        pattern_matching_result_1 = 1
                                        t2_7 = tail(l2)
                                        x2_5 = head(l2)

                                    else:
                                        pattern_matching_result_1 = 2

                                else:
                                    pattern_matching_result_1 = 2

                                if pattern_matching_result_1 == 0:
                                    if isinstance(x1_5, SetTreeNode_1):
                                        comparer_mut = comparer
                                        l1_mut = of_array_with_tail(
                                            [
                                                SetTreeNode_1__get_Left(x1_5),
                                                SetTreeNode_1__ctor_5F465FC9(
                                                    SetTreeLeaf_1__get_Key(x1_5),
                                                    SetTreeModule_empty(),
                                                    SetTreeNode_1__get_Right(x1_5),
                                                    0,
                                                ),
                                            ],
                                            t1_7,
                                        )
                                        l2_mut = l2
                                        continue

                                    else:
                                        comparer_mut = comparer
                                        l1_mut = of_array_with_tail(
                                            [
                                                SetTreeModule_empty(),
                                                SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x1_5)),
                                            ],
                                            t1_7,
                                        )
                                        l2_mut = l2
                                        continue

                                elif pattern_matching_result_1 == 1:
                                    if isinstance(x2_5, SetTreeNode_1):
                                        comparer_mut = comparer
                                        l1_mut = l1
                                        l2_mut = of_array_with_tail(
                                            [
                                                SetTreeNode_1__get_Left(x2_5),
                                                SetTreeNode_1__ctor_5F465FC9(
                                                    SetTreeLeaf_1__get_Key(x2_5),
                                                    SetTreeModule_empty(),
                                                    SetTreeNode_1__get_Right(x2_5),
                                                    0,
                                                ),
                                            ],
                                            t2_7,
                                        )
                                        continue

                                    else:
                                        comparer_mut = comparer
                                        l1_mut = l1
                                        l2_mut = of_array_with_tail(
                                            [
                                                SetTreeModule_empty(),
                                                SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2_5)),
                                            ],
                                            t2_7,
                                        )
                                        continue

                                elif pattern_matching_result_1 == 2:
                                    raise Exception("unexpected state in SetTree.compareStacks")

                        elif isinstance(x2_3, SetTreeNode_1):
                            if SetTreeNode_1__get_Left(x2_3) is None:
                                c_2: int = (
                                    comparer.Compare(SetTreeLeaf_1__get_Key(x1_3), SetTreeLeaf_1__get_Key(x2_3)) or 0
                                )
                                if c_2 != 0:
                                    return c_2

                                else:
                                    comparer_mut = comparer
                                    l1_mut = cons(SetTreeModule_empty(), tail(l1))
                                    l2_mut = cons(SetTreeNode_1__get_Right(x2_3), tail(l2))
                                    continue

                            else:
                                (pattern_matching_result_2, t1_8, x1_6, t2_8, x2_6) = (None, None, None, None, None)
                                if not is_empty_1(l1):
                                    if head(l1) is not None:
                                        pattern_matching_result_2 = 0
                                        t1_8 = tail(l1)
                                        x1_6 = head(l1)

                                    elif not is_empty_1(l2):
                                        if head(l2) is not None:
                                            pattern_matching_result_2 = 1
                                            t2_8 = tail(l2)
                                            x2_6 = head(l2)

                                        else:
                                            pattern_matching_result_2 = 2

                                    else:
                                        pattern_matching_result_2 = 2

                                elif not is_empty_1(l2):
                                    if head(l2) is not None:
                                        pattern_matching_result_2 = 1
                                        t2_8 = tail(l2)
                                        x2_6 = head(l2)

                                    else:
                                        pattern_matching_result_2 = 2

                                else:
                                    pattern_matching_result_2 = 2

                                if pattern_matching_result_2 == 0:
                                    if isinstance(x1_6, SetTreeNode_1):
                                        comparer_mut = comparer
                                        l1_mut = of_array_with_tail(
                                            [
                                                SetTreeNode_1__get_Left(x1_6),
                                                SetTreeNode_1__ctor_5F465FC9(
                                                    SetTreeLeaf_1__get_Key(x1_6),
                                                    SetTreeModule_empty(),
                                                    SetTreeNode_1__get_Right(x1_6),
                                                    0,
                                                ),
                                            ],
                                            t1_8,
                                        )
                                        l2_mut = l2
                                        continue

                                    else:
                                        comparer_mut = comparer
                                        l1_mut = of_array_with_tail(
                                            [
                                                SetTreeModule_empty(),
                                                SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x1_6)),
                                            ],
                                            t1_8,
                                        )
                                        l2_mut = l2
                                        continue

                                elif pattern_matching_result_2 == 1:
                                    if isinstance(x2_6, SetTreeNode_1):
                                        comparer_mut = comparer
                                        l1_mut = l1
                                        l2_mut = of_array_with_tail(
                                            [
                                                SetTreeNode_1__get_Left(x2_6),
                                                SetTreeNode_1__ctor_5F465FC9(
                                                    SetTreeLeaf_1__get_Key(x2_6),
                                                    SetTreeModule_empty(),
                                                    SetTreeNode_1__get_Right(x2_6),
                                                    0,
                                                ),
                                            ],
                                            t2_8,
                                        )
                                        continue

                                    else:
                                        comparer_mut = comparer
                                        l1_mut = l1
                                        l2_mut = of_array_with_tail(
                                            [
                                                SetTreeModule_empty(),
                                                SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2_6)),
                                            ],
                                            t2_8,
                                        )
                                        continue

                                elif pattern_matching_result_2 == 2:
                                    raise Exception("unexpected state in SetTree.compareStacks")

                        else:
                            c_3: int = comparer.Compare(SetTreeLeaf_1__get_Key(x1_3), SetTreeLeaf_1__get_Key(x2_3)) or 0
                            if c_3 != 0:
                                return c_3

                            else:
                                comparer_mut = comparer
                                l1_mut = tail(l1)
                                l2_mut = tail(l2)
                                continue

                    else:
                        x2: SetTreeLeaf_1[_T] = head(l2)
                        (pattern_matching_result_3, t1_2, x1, t2_2, x2_1) = (None, None, None, None, None)
                        if not is_empty_1(l1):
                            if head(l1) is not None:
                                pattern_matching_result_3 = 0
                                t1_2 = tail(l1)
                                x1 = head(l1)

                            elif not is_empty_1(l2):
                                if head(l2) is not None:
                                    pattern_matching_result_3 = 1
                                    t2_2 = tail(l2)
                                    x2_1 = head(l2)

                                else:
                                    pattern_matching_result_3 = 2

                            else:
                                pattern_matching_result_3 = 2

                        elif not is_empty_1(l2):
                            if head(l2) is not None:
                                pattern_matching_result_3 = 1
                                t2_2 = tail(l2)
                                x2_1 = head(l2)

                            else:
                                pattern_matching_result_3 = 2

                        else:
                            pattern_matching_result_3 = 2

                        if pattern_matching_result_3 == 0:
                            if isinstance(x1, SetTreeNode_1):
                                comparer_mut = comparer
                                l1_mut = of_array_with_tail(
                                    [
                                        SetTreeNode_1__get_Left(x1),
                                        SetTreeNode_1__ctor_5F465FC9(
                                            SetTreeLeaf_1__get_Key(x1),
                                            SetTreeModule_empty(),
                                            SetTreeNode_1__get_Right(x1),
                                            0,
                                        ),
                                    ],
                                    t1_2,
                                )
                                l2_mut = l2
                                continue

                            else:
                                comparer_mut = comparer
                                l1_mut = of_array_with_tail(
                                    [SetTreeModule_empty(), SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x1))], t1_2
                                )
                                l2_mut = l2
                                continue

                        elif pattern_matching_result_3 == 1:
                            if isinstance(x2_1, SetTreeNode_1):
                                comparer_mut = comparer
                                l1_mut = l1
                                l2_mut = of_array_with_tail(
                                    [
                                        SetTreeNode_1__get_Left(x2_1),
                                        SetTreeNode_1__ctor_5F465FC9(
                                            SetTreeLeaf_1__get_Key(x2_1),
                                            SetTreeModule_empty(),
                                            SetTreeNode_1__get_Right(x2_1),
                                            0,
                                        ),
                                    ],
                                    t2_2,
                                )
                                continue

                            else:
                                comparer_mut = comparer
                                l1_mut = l1
                                l2_mut = of_array_with_tail(
                                    [SetTreeModule_empty(), SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2_1))],
                                    t2_2,
                                )
                                continue

                        elif pattern_matching_result_3 == 2:
                            raise Exception("unexpected state in SetTree.compareStacks")

                elif head(l1) is not None:
                    x1_1: SetTreeLeaf_1[_T] = head(l1)
                    (pattern_matching_result_4, t1_4, x1_2, t2_4, x2_2) = (None, None, None, None, None)
                    if not is_empty_1(l1):
                        if head(l1) is not None:
                            pattern_matching_result_4 = 0
                            t1_4 = tail(l1)
                            x1_2 = head(l1)

                        elif not is_empty_1(l2):
                            if head(l2) is not None:
                                pattern_matching_result_4 = 1
                                t2_4 = tail(l2)
                                x2_2 = head(l2)

                            else:
                                pattern_matching_result_4 = 2

                        else:
                            pattern_matching_result_4 = 2

                    elif not is_empty_1(l2):
                        if head(l2) is not None:
                            pattern_matching_result_4 = 1
                            t2_4 = tail(l2)
                            x2_2 = head(l2)

                        else:
                            pattern_matching_result_4 = 2

                    else:
                        pattern_matching_result_4 = 2

                    if pattern_matching_result_4 == 0:
                        if isinstance(x1_2, SetTreeNode_1):
                            comparer_mut = comparer
                            l1_mut = of_array_with_tail(
                                [
                                    SetTreeNode_1__get_Left(x1_2),
                                    SetTreeNode_1__ctor_5F465FC9(
                                        SetTreeLeaf_1__get_Key(x1_2),
                                        SetTreeModule_empty(),
                                        SetTreeNode_1__get_Right(x1_2),
                                        0,
                                    ),
                                ],
                                t1_4,
                            )
                            l2_mut = l2
                            continue

                        else:
                            comparer_mut = comparer
                            l1_mut = of_array_with_tail(
                                [SetTreeModule_empty(), SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x1_2))], t1_4
                            )
                            l2_mut = l2
                            continue

                    elif pattern_matching_result_4 == 1:
                        if isinstance(x2_2, SetTreeNode_1):
                            comparer_mut = comparer
                            l1_mut = l1
                            l2_mut = of_array_with_tail(
                                [
                                    SetTreeNode_1__get_Left(x2_2),
                                    SetTreeNode_1__ctor_5F465FC9(
                                        SetTreeLeaf_1__get_Key(x2_2),
                                        SetTreeModule_empty(),
                                        SetTreeNode_1__get_Right(x2_2),
                                        0,
                                    ),
                                ],
                                t2_4,
                            )
                            continue

                        else:
                            comparer_mut = comparer
                            l1_mut = l1
                            l2_mut = of_array_with_tail(
                                [SetTreeModule_empty(), SetTreeLeaf_1__ctor_2B595(SetTreeLeaf_1__get_Key(x2_2))], t2_4
                            )
                            continue

                    elif pattern_matching_result_4 == 2:
                        raise Exception("unexpected state in SetTree.compareStacks")

                else:
                    comparer_mut = comparer
                    l1_mut = tail(l1)
                    l2_mut = tail(l2)
                    continue

            else:
                return 1

        elif is_empty_1(l2):
            return 0

        else:
            return -1

        break


def SetTreeModule_compare(
    comparer: IComparer_1[Any], t1: SetTreeLeaf_1[Any] | None = None, t2: SetTreeLeaf_1[Any] | None = None
) -> int:
    if t1 is None:
        if t2 is None:
            return 0

        else:
            return -1

    elif t2 is None:
        return 1

    else:
        return SetTreeModule_compareStacks(comparer, singleton_1(t1), singleton_1(t2))


def SetTreeModule_choose(s: SetTreeLeaf_1[Any] | None = None) -> Any:
    return SetTreeModule_minimumElement(s)


def SetTreeModule_toList(t: SetTreeLeaf_1[Any] | None = None) -> FSharpList[Any]:
    def loop(t_0027_mut: SetTreeLeaf_1[_T] | None, acc_mut: FSharpList[_T], t: Any = t) -> FSharpList[_T]:
        while True:
            (t_0027, acc) = (t_0027_mut, acc_mut)
            if t_0027 is not None:
                t2: SetTreeLeaf_1[_T] = t_0027
                if isinstance(t2, SetTreeNode_1):
                    t_0027_mut = SetTreeNode_1__get_Left(t2)
                    acc_mut = cons(SetTreeLeaf_1__get_Key(t2), loop(SetTreeNode_1__get_Right(t2), acc))
                    continue

                else:
                    return cons(SetTreeLeaf_1__get_Key(t2), acc)

            else:
                return acc

            break

    return loop(t, empty_1())


def SetTreeModule_copyToArray(s: SetTreeLeaf_1[Any] | None, arr: Array[Any], i: int) -> None:
    j: int = i or 0

    def _arrow240(x: __A | None = None, s: Any = s, arr: Any = arr, i: Any = i) -> None:
        nonlocal j
        arr[j] = x
        j = (j + 1) or 0

    SetTreeModule_iter(_arrow240, s)


def SetTreeModule_toArray(s: SetTreeLeaf_1[Any] | None = None) -> Array[Any]:
    n: int = SetTreeModule_count(s) or 0
    res: Array[__A] = fill([0] * n, 0, n, None)
    SetTreeModule_copyToArray(s, res, 0)
    return res


def SetTreeModule_mkFromEnumerator(
    comparer_mut: IComparer_1[Any], acc_mut: SetTreeLeaf_1[Any] | None, e_mut: IEnumerator[Any]
) -> SetTreeLeaf_1[Any] | None:
    while True:
        (comparer, acc, e) = (comparer_mut, acc_mut, e_mut)
        if e.System_Collections_IEnumerator_MoveNext():
            comparer_mut = comparer
            acc_mut = SetTreeModule_add(comparer, e.System_Collections_Generic_IEnumerator_1_get_Current(), acc)
            e_mut = e
            continue

        else:
            return acc

        break


def SetTreeModule_ofArray(comparer: IComparer_1[Any], l: Array[Any]) -> SetTreeLeaf_1[Any] | None:
    def _arrow241(
        acc: SetTreeLeaf_1[__A] | None, k: __A, comparer: Any = comparer, l: Any = l
    ) -> SetTreeLeaf_1[__A] | None:
        return SetTreeModule_add(comparer, k, acc)

    return fold_1(_arrow241, SetTreeModule_empty(), l)


def SetTreeModule_ofList(comparer: IComparer_1[Any], l: FSharpList[Any]) -> SetTreeLeaf_1[Any] | None:
    def _arrow242(
        acc: SetTreeLeaf_1[__A] | None, k: __A, comparer: Any = comparer, l: Any = l
    ) -> SetTreeLeaf_1[__A] | None:
        return SetTreeModule_add(comparer, k, acc)

    return fold_2(_arrow242, SetTreeModule_empty(), l)


def SetTreeModule_ofSeq(comparer: IComparer_1[Any], c: IEnumerable_1[Any]) -> SetTreeLeaf_1[Any] | None:
    if is_array_like(c):
        return SetTreeModule_ofArray(comparer, c)

    elif isinstance(c, FSharpList):
        return SetTreeModule_ofList(comparer, c)

    else:
        with get_enumerator(c) as ie:
            return SetTreeModule_mkFromEnumerator(comparer, SetTreeModule_empty(), ie)


def _expr244(gen0: TypeInfo) -> TypeInfo:
    return class_type("Set.FSharpSet", [gen0], FSharpSet)


class FSharpSet(Generic[_T]):
    def __init__(self, comparer: IComparer_1[Any], tree: SetTreeLeaf_1[Any] | None = None) -> None:
        self.comparer: IComparer_1[_T] = comparer
        self.tree: SetTreeLeaf_1[_T] | None = tree

    def __hash__(self, __unit: None = None) -> int:
        this: FSharpSet[_T] = self
        return FSharpSet__ComputeHashCode(this)

    def __eq__(self, that: Any = None) -> bool:
        this: FSharpSet[_T] = self
        return (
            (
                SetTreeModule_compare(
                    FSharpSet__get_Comparer(this), FSharpSet__get_Tree(this), FSharpSet__get_Tree(that)
                )
                == 0
            )
            if isinstance(that, FSharpSet)
            else False
        )

    def __str__(self, __unit: None = None) -> str:
        this: FSharpSet[_T] = self

        def _arrow243(x: _T | None = None) -> str:
            copy_of_struct: _T = x
            return to_string(copy_of_struct)

        return ("set [" + join("; ", map_1(_arrow243, this))) + "]"

    @property
    def Symbol_toStringTag(self, __unit: None = None) -> str:
        return "FSharpSet"

    def to_json(self, __unit: None = None) -> Any:
        this: FSharpSet[_T] = self
        return list(this)

    def __cmp__(self, that: Any = None) -> int:
        s: FSharpSet[_T] = self
        return SetTreeModule_compare(FSharpSet__get_Comparer(s), FSharpSet__get_Tree(s), FSharpSet__get_Tree(that))

    def System_Collections_Generic_ICollection_1_Add2B595(self, x: _T | None = None) -> None:
        ignore(x)
        raise Exception("ReadOnlyCollection")

    def System_Collections_Generic_ICollection_1_Clear(self, __unit: None = None) -> None:
        raise Exception("ReadOnlyCollection")

    def System_Collections_Generic_ICollection_1_Remove2B595(self, x: _T | None = None) -> bool:
        ignore(x)
        raise Exception("ReadOnlyCollection")

    def System_Collections_Generic_ICollection_1_Contains2B595(self, x: _T | None = None) -> bool:
        s: FSharpSet[_T] = self
        return SetTreeModule_mem(FSharpSet__get_Comparer(s), x, FSharpSet__get_Tree(s))

    def System_Collections_Generic_ICollection_1_CopyToZ3B4C077E(self, arr: Array[_T], i: int) -> None:
        s: FSharpSet[_T] = self
        SetTreeModule_copyToArray(FSharpSet__get_Tree(s), arr, i)

    def System_Collections_Generic_ICollection_1_get_IsReadOnly(self, __unit: None = None) -> bool:
        return True

    def __len__(self, __unit: None = None) -> int:
        s: FSharpSet[_T] = self
        return FSharpSet__get_Count(s)

    def __len__(self, __unit: None = None) -> int:
        s: FSharpSet[_T] = self
        return FSharpSet__get_Count(s)

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        s: FSharpSet[_T] = self
        return SetTreeModule_mkIEnumerator(FSharpSet__get_Tree(s))

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        s: FSharpSet[_T] = self
        return SetTreeModule_mkIEnumerator(FSharpSet__get_Tree(s))

    @property
    def size(self, __unit: None = None) -> int:
        s: FSharpSet[_T] = self
        return FSharpSet__get_Count(s)

    def add(self, k: _T | None = None) -> Set_1[_T]:
        s: FSharpSet[_T] = self
        raise Exception("Set cannot be mutated")
        return s

    def clear(self, __unit: None = None) -> None:
        raise Exception("Set cannot be mutated")

    def __delitem__(self, k: _T | None = None) -> bool:
        raise Exception("Set cannot be mutated")
        return False

    def __contains__(self, k: _T | None = None) -> bool:
        s: FSharpSet[_T] = self
        return FSharpSet__Contains(s, k)

    def keys(self, __unit: None = None) -> IEnumerable_1[_T]:
        s: FSharpSet[_T] = self

        def mapping(x: _T | None = None) -> _T | None:
            return x

        return map_1(mapping, s)

    def values(self, __unit: None = None) -> IEnumerable_1[_T]:
        s: FSharpSet[_T] = self

        def mapping(x: _T | None = None) -> _T | None:
            return x

        return map_1(mapping, s)

    def entries(self, __unit: None = None) -> IEnumerable_1[tuple[_T, _T]]:
        s: FSharpSet[_T] = self

        def mapping(v: _T | None = None) -> tuple[_T, _T]:
            return (v, v)

        return map_1(mapping, s)

    def for_each(self, f: Callable[[_T, _T, Set_1[_T]], None], this_arg: Any | None = None) -> None:
        s: FSharpSet[_T] = self

        def action(x: _T | None = None) -> None:
            f(x, x, s)

        iterate_1(action, s)


FSharpSet_reflection = _expr244


def FSharpSet__ctor(comparer: IComparer_1[Any], tree: SetTreeLeaf_1[Any] | None = None) -> FSharpSet[_T]:
    return FSharpSet(comparer, tree)


def FSharpSet__get_Comparer(set_1: FSharpSet[Any]) -> IComparer_1[Any]:
    return set_1.comparer


def FSharpSet__get_Tree(set_1: FSharpSet[Any]) -> SetTreeLeaf_1[Any] | None:
    return set_1.tree


def FSharpSet_Empty(comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__ctor(comparer, SetTreeModule_empty())


def FSharpSet__Add(s: FSharpSet[Any], value: Any) -> FSharpSet[Any]:
    return FSharpSet__ctor(
        FSharpSet__get_Comparer(s), SetTreeModule_add(FSharpSet__get_Comparer(s), value, FSharpSet__get_Tree(s))
    )


def FSharpSet__Remove(s: FSharpSet[Any], value: Any) -> FSharpSet[Any]:
    return FSharpSet__ctor(
        FSharpSet__get_Comparer(s), SetTreeModule_remove(FSharpSet__get_Comparer(s), value, FSharpSet__get_Tree(s))
    )


def FSharpSet__get_Count(s: FSharpSet[Any]) -> int:
    return SetTreeModule_count(FSharpSet__get_Tree(s))


def FSharpSet__Contains(s: FSharpSet[Any], value: Any) -> bool:
    return SetTreeModule_mem(FSharpSet__get_Comparer(s), value, FSharpSet__get_Tree(s))


def FSharpSet__Iterate(s: FSharpSet[Any], x: Callable[[_T], None]) -> None:
    SetTreeModule_iter(x, FSharpSet__get_Tree(s))


def FSharpSet__Fold(s: FSharpSet[Any], f: Callable[[_T, __A], __A], z: Any) -> Any:
    f_1: Any = f

    def _arrow245(x: __A, z_1: _T, s: Any = s, f: Any = f, z: Any = z) -> __A:
        return f_1(z_1, x)

    return SetTreeModule_fold(_arrow245, z, FSharpSet__get_Tree(s))


def FSharpSet__get_IsEmpty(s: FSharpSet[Any]) -> bool:
    return FSharpSet__get_Tree(s) is None


def FSharpSet__Partition(s: FSharpSet[Any], f: Callable[[_T], bool]) -> tuple[FSharpSet[_T], FSharpSet[_T]]:
    if FSharpSet__get_Tree(s) is None:
        return (s, s)

    else:
        pattern_input: tuple[SetTreeLeaf_1[_T] | None, SetTreeLeaf_1[_T] | None] = SetTreeModule_partition(
            FSharpSet__get_Comparer(s), f, FSharpSet__get_Tree(s)
        )
        return (
            FSharpSet__ctor(FSharpSet__get_Comparer(s), pattern_input[0]),
            FSharpSet__ctor(FSharpSet__get_Comparer(s), pattern_input[1]),
        )


def FSharpSet__Filter(s: FSharpSet[Any], f: Callable[[_T], bool]) -> FSharpSet[Any]:
    if FSharpSet__get_Tree(s) is None:
        return s

    else:
        return FSharpSet__ctor(
            FSharpSet__get_Comparer(s), SetTreeModule_filter(FSharpSet__get_Comparer(s), f, FSharpSet__get_Tree(s))
        )


def FSharpSet__Map(s: FSharpSet[Any], f: Callable[[_T], _U], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    def _arrow246(
        acc: SetTreeLeaf_1[_U] | None, k: _T, s: Any = s, f: Any = f, comparer: Any = comparer
    ) -> SetTreeLeaf_1[_U] | None:
        return SetTreeModule_add(comparer, f(k), acc)

    return FSharpSet__ctor(comparer, SetTreeModule_fold(_arrow246, SetTreeModule_empty(), FSharpSet__get_Tree(s)))


def FSharpSet__Exists(s: FSharpSet[Any], f: Callable[[_T], bool]) -> bool:
    return SetTreeModule_exists(f, FSharpSet__get_Tree(s))


def FSharpSet__ForAll(s: FSharpSet[Any], f: Callable[[_T], bool]) -> bool:
    return SetTreeModule_forall(f, FSharpSet__get_Tree(s))


def FSharpSet_op_Subtraction(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> FSharpSet[Any]:
    if FSharpSet__get_Tree(set1) is None:
        return set1

    elif FSharpSet__get_Tree(set2) is None:
        return set1

    else:
        return FSharpSet__ctor(
            FSharpSet__get_Comparer(set1),
            SetTreeModule_diff(FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set1), FSharpSet__get_Tree(set2)),
        )


def FSharpSet_op_Addition(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> FSharpSet[Any]:
    if FSharpSet__get_Tree(set2) is None:
        return set1

    elif FSharpSet__get_Tree(set1) is None:
        return set2

    else:
        return FSharpSet__ctor(
            FSharpSet__get_Comparer(set1),
            SetTreeModule_union(FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set1), FSharpSet__get_Tree(set2)),
        )


def FSharpSet_Intersection(a: FSharpSet[Any], b: FSharpSet[Any]) -> FSharpSet[Any]:
    if FSharpSet__get_Tree(b) is None:
        return b

    elif FSharpSet__get_Tree(a) is None:
        return a

    else:
        return FSharpSet__ctor(
            FSharpSet__get_Comparer(a),
            SetTreeModule_intersection(FSharpSet__get_Comparer(a), FSharpSet__get_Tree(a), FSharpSet__get_Tree(b)),
        )


def FSharpSet_IntersectionMany(sets: IEnumerable_1[FSharpSet[Any]]) -> FSharpSet[Any]:
    def _arrow247(s1: FSharpSet[_T], s2: FSharpSet[_T], sets: Any = sets) -> FSharpSet[_T]:
        return FSharpSet_Intersection(s1, s2)

    return reduce(_arrow247, sets)


def FSharpSet_Equality(a: FSharpSet[Any], b: FSharpSet[Any]) -> bool:
    return SetTreeModule_compare(FSharpSet__get_Comparer(a), FSharpSet__get_Tree(a), FSharpSet__get_Tree(b)) == 0


def FSharpSet_Compare(a: FSharpSet[Any], b: FSharpSet[Any]) -> int:
    return SetTreeModule_compare(FSharpSet__get_Comparer(a), FSharpSet__get_Tree(a), FSharpSet__get_Tree(b))


def FSharpSet__get_Choose(x: FSharpSet[Any]) -> Any:
    return SetTreeModule_choose(FSharpSet__get_Tree(x))


def FSharpSet__get_MinimumElement(x: FSharpSet[Any]) -> Any:
    return SetTreeModule_minimumElement(FSharpSet__get_Tree(x))


def FSharpSet__get_MaximumElement(x: FSharpSet[Any]) -> Any:
    return SetTreeModule_maximumElement(FSharpSet__get_Tree(x))


def FSharpSet__IsSubsetOf(x: FSharpSet[Any], other_set: FSharpSet[Any]) -> bool:
    return SetTreeModule_subset(FSharpSet__get_Comparer(x), FSharpSet__get_Tree(x), FSharpSet__get_Tree(other_set))


def FSharpSet__IsSupersetOf(x: FSharpSet[Any], other_set: FSharpSet[Any]) -> bool:
    return SetTreeModule_subset(FSharpSet__get_Comparer(x), FSharpSet__get_Tree(other_set), FSharpSet__get_Tree(x))


def FSharpSet__IsProperSubsetOf(x: FSharpSet[Any], other_set: FSharpSet[Any]) -> bool:
    return SetTreeModule_properSubset(
        FSharpSet__get_Comparer(x), FSharpSet__get_Tree(x), FSharpSet__get_Tree(other_set)
    )


def FSharpSet__IsProperSupersetOf(x: FSharpSet[Any], other_set: FSharpSet[Any]) -> bool:
    return SetTreeModule_properSubset(
        FSharpSet__get_Comparer(x), FSharpSet__get_Tree(other_set), FSharpSet__get_Tree(x)
    )


def FSharpSet__ToList(x: FSharpSet[Any]) -> FSharpList[Any]:
    return SetTreeModule_toList(FSharpSet__get_Tree(x))


def FSharpSet__ToArray(x: FSharpSet[Any]) -> Array[Any]:
    return SetTreeModule_toArray(FSharpSet__get_Tree(x))


def FSharpSet__ComputeHashCode(this: FSharpSet[Any]) -> int:
    res: int = 0
    with get_enumerator(this) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            x_1: _T = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            res = (((res << 1) + structural_hash(x_1)) + 631) or 0
    return abs(res)


def is_empty(set_1: FSharpSet[Any]) -> bool:
    return FSharpSet__get_IsEmpty(set_1)


def contains(element: Any, set_1: FSharpSet[Any]) -> bool:
    return FSharpSet__Contains(set_1, element)


def add(value: Any, set_1: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet__Add(set_1, value)


def singleton(value: Any, comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__Add(FSharpSet_Empty(comparer), value)


def remove(value: Any, set_1: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet__Remove(set_1, value)


def union(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet_op_Addition(set1, set2)


def union_many(sets: IEnumerable_1[FSharpSet[Any]], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    def _arrow248(x: FSharpSet[_T], y: FSharpSet[_T], sets: Any = sets, comparer: Any = comparer) -> FSharpSet[_T]:
        return FSharpSet_op_Addition(x, y)

    return fold_3(_arrow248, FSharpSet_Empty(comparer), sets)


def intersect(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet_Intersection(set1, set2)


def intersect_many(sets: IEnumerable_1[FSharpSet[Any]]) -> FSharpSet[Any]:
    return FSharpSet_IntersectionMany(sets)


def iterate(action: Callable[[_T], None], set_1: FSharpSet[Any]) -> None:
    FSharpSet__Iterate(set_1, action)


def empty(comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet_Empty(comparer)


def for_all(predicate: Callable[[_T], bool], set_1: FSharpSet[Any]) -> bool:
    return FSharpSet__ForAll(set_1, predicate)


def exists(predicate: Callable[[_T], bool], set_1: FSharpSet[Any]) -> bool:
    return FSharpSet__Exists(set_1, predicate)


def filter(predicate: Callable[[_T], bool], set_1: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet__Filter(set_1, predicate)


def partition(predicate: Callable[[_T], bool], set_1: FSharpSet[Any]) -> tuple[FSharpSet[_T], FSharpSet[_T]]:
    return FSharpSet__Partition(set_1, predicate)


def fold(folder: Callable[[_STATE, _T], _STATE], state: Any, set_1: FSharpSet[Any]) -> Any:
    return SetTreeModule_fold(folder, state, FSharpSet__get_Tree(set_1))


def fold_back(folder: Callable[[_T, _STATE], _STATE], set_1: FSharpSet[Any], state: Any) -> Any:
    return SetTreeModule_foldBack(folder, FSharpSet__get_Tree(set_1), state)


def map(mapping: Callable[[_T], _U], set_1: FSharpSet[Any], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__Map(set_1, mapping, comparer)


def count(set_1: FSharpSet[Any]) -> int:
    return FSharpSet__get_Count(set_1)


def of_list(elements: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__ctor(comparer, SetTreeModule_ofSeq(comparer, elements))


def of_array(array: Array[Any], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__ctor(comparer, SetTreeModule_ofArray(comparer, array))


def to_list(set_1: FSharpSet[Any]) -> FSharpList[Any]:
    return FSharpSet__ToList(set_1)


def to_array(set_1: FSharpSet[Any]) -> Array[Any]:
    return FSharpSet__ToArray(set_1)


def to_seq(set_1: FSharpSet[Any]) -> IEnumerable_1[Any]:
    def mapping(x: _T | None = None, set_1: Any = set_1) -> _T | None:
        return x

    return map_1(mapping, set_1)


def of_seq(elements: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> FSharpSet[Any]:
    return FSharpSet__ctor(comparer, SetTreeModule_ofSeq(comparer, elements))


def difference(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> FSharpSet[Any]:
    return FSharpSet_op_Subtraction(set1, set2)


def is_subset(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> bool:
    return SetTreeModule_subset(FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set1), FSharpSet__get_Tree(set2))


def is_superset(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> bool:
    return SetTreeModule_subset(FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set2), FSharpSet__get_Tree(set1))


def is_proper_subset(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> bool:
    return SetTreeModule_properSubset(
        FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set1), FSharpSet__get_Tree(set2)
    )


def is_proper_superset(set1: FSharpSet[Any], set2: FSharpSet[Any]) -> bool:
    return SetTreeModule_properSubset(
        FSharpSet__get_Comparer(set1), FSharpSet__get_Tree(set2), FSharpSet__get_Tree(set1)
    )


def min_element(set_1: FSharpSet[Any]) -> Any:
    return FSharpSet__get_MinimumElement(set_1)


def max_element(set_1: FSharpSet[Any]) -> Any:
    return FSharpSet__get_MaximumElement(set_1)


def union_with(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> Set_1[Any]:
    def folder(acc: Set_1[_T], x: _T, s1: Any = s1, s2: Any = s2) -> Set_1[_T]:
        return acc.add(x)

    return fold_3(folder, s1, s2)


def new_mutable_set_with(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> Set_1[Any]:
    if isinstance(s1, HashSet):
        return HashSet__ctor_Z6150332D(s2, HashSet__get_Comparer(s1))

    else:
        return Set(s2)


def intersect_with(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> None:
    s2_1: Set_1[_T] = new_mutable_set_with(s1, s2)

    def action(x: _T | None = None, s1: Any = s1, s2: Any = s2) -> None:
        if not (x in s2_1):
            ignore(s1.delete(x))

    iterate_1(action, s1.values())


def except_with(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> None:
    def action(x: _T | None = None, s1: Any = s1, s2: Any = s2) -> None:
        ignore(s1.delete(x))

    iterate_1(action, s2)


def is_subset_of(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> bool:
    s2_1: Set_1[_T] = new_mutable_set_with(s1, s2)

    def predicate(value: _T | None = None, s1: Any = s1, s2: Any = s2) -> bool:
        return value in s2_1

    return for_all_1(predicate, s1.values())


def is_superset_of(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> bool:
    def predicate(value: _T | None = None, s1: Any = s1, s2: Any = s2) -> bool:
        return value in s1

    return for_all_1(predicate, s2)


def is_proper_subset_of(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> bool:
    s2_1: Set_1[_T] = new_mutable_set_with(s1, s2)
    if s2_1.size > s1.size:

        def predicate(value: _T | None = None, s1: Any = s1, s2: Any = s2) -> bool:
            return value in s2_1

        return for_all_1(predicate, s1.values())

    else:
        return False


def is_proper_superset_of(s1: Set_1[Any], s2: IEnumerable_1[Any]) -> bool:
    s2_1: IEnumerable_1[_T] = cache(s2)

    def predicate(arg: _T | None = None, s1: Any = s1, s2: Any = s2) -> bool:
        return not (arg in s1)

    if exists_1(predicate, s2_1):

        def predicate_1(value_2: _T | None = None, s1: Any = s1, s2: Any = s2) -> bool:
            return value_2 in s1

        return for_all_1(predicate_1, s2_1)

    else:
        return False


__all__ = [
    "SetTreeLeaf_1_reflection",
    "SetTreeLeaf_1__get_Key",
    "SetTreeNode_1_reflection",
    "SetTreeNode_1__get_Left",
    "SetTreeNode_1__get_Right",
    "SetTreeNode_1__get_Height",
    "SetTreeModule_empty",
    "SetTreeModule_countAux",
    "SetTreeModule_count",
    "SetTreeModule_mk",
    "SetTreeModule_rebalance",
    "SetTreeModule_add",
    "SetTreeModule_balance",
    "SetTreeModule_split",
    "SetTreeModule_spliceOutSuccessor",
    "SetTreeModule_remove",
    "SetTreeModule_mem",
    "SetTreeModule_iter",
    "SetTreeModule_foldBackOpt",
    "SetTreeModule_foldBack",
    "SetTreeModule_foldOpt",
    "SetTreeModule_fold",
    "SetTreeModule_forall",
    "SetTreeModule_exists",
    "SetTreeModule_subset",
    "SetTreeModule_properSubset",
    "SetTreeModule_filterAux",
    "SetTreeModule_filter",
    "SetTreeModule_diffAux",
    "SetTreeModule_diff",
    "SetTreeModule_union",
    "SetTreeModule_intersectionAux",
    "SetTreeModule_intersection",
    "SetTreeModule_partition1",
    "SetTreeModule_partitionAux",
    "SetTreeModule_partition",
    "SetTreeModule_minimumElementAux",
    "SetTreeModule_minimumElementOpt",
    "SetTreeModule_maximumElementAux",
    "SetTreeModule_maximumElementOpt",
    "SetTreeModule_minimumElement",
    "SetTreeModule_maximumElement",
    "SetTreeModule_SetIterator_1_reflection",
    "SetTreeModule_collapseLHS",
    "SetTreeModule_mkIterator",
    "SetTreeModule_notStarted",
    "SetTreeModule_alreadyFinished",
    "SetTreeModule_current",
    "SetTreeModule_moveNext",
    "SetTreeModule_mkIEnumerator",
    "SetTreeModule_compareStacks",
    "SetTreeModule_compare",
    "SetTreeModule_choose",
    "SetTreeModule_toList",
    "SetTreeModule_copyToArray",
    "SetTreeModule_toArray",
    "SetTreeModule_mkFromEnumerator",
    "SetTreeModule_ofArray",
    "SetTreeModule_ofList",
    "SetTreeModule_ofSeq",
    "FSharpSet_reflection",
    "FSharpSet__get_Comparer",
    "FSharpSet__get_Tree",
    "FSharpSet_Empty",
    "FSharpSet__Add",
    "FSharpSet__Remove",
    "FSharpSet__get_Count",
    "FSharpSet__Contains",
    "FSharpSet__Iterate",
    "FSharpSet__Fold",
    "FSharpSet__get_IsEmpty",
    "FSharpSet__Partition",
    "FSharpSet__Filter",
    "FSharpSet__Map",
    "FSharpSet__Exists",
    "FSharpSet__ForAll",
    "FSharpSet_op_Subtraction",
    "FSharpSet_op_Addition",
    "FSharpSet_Intersection",
    "FSharpSet_IntersectionMany",
    "FSharpSet_Equality",
    "FSharpSet_Compare",
    "FSharpSet__get_Choose",
    "FSharpSet__get_MinimumElement",
    "FSharpSet__get_MaximumElement",
    "FSharpSet__IsSubsetOf",
    "FSharpSet__IsSupersetOf",
    "FSharpSet__IsProperSubsetOf",
    "FSharpSet__IsProperSupersetOf",
    "FSharpSet__ToList",
    "FSharpSet__ToArray",
    "FSharpSet__ComputeHashCode",
    "is_empty",
    "contains",
    "add",
    "singleton",
    "remove",
    "union",
    "union_many",
    "intersect",
    "intersect_many",
    "iterate",
    "empty",
    "for_all",
    "exists",
    "filter",
    "partition",
    "fold",
    "fold_back",
    "map",
    "count",
    "of_list",
    "of_array",
    "to_list",
    "to_array",
    "to_seq",
    "of_seq",
    "difference",
    "is_subset",
    "is_superset",
    "is_proper_subset",
    "is_proper_superset",
    "min_element",
    "max_element",
    "union_with",
    "new_mutable_set_with",
    "intersect_with",
    "except_with",
    "is_subset_of",
    "is_superset_of",
    "is_proper_subset_of",
    "is_proper_superset_of",
]



from __future__ import annotations

from collections.abc import Callable
from decimal import Decimal
from typing import Any, TypeVar

from .big_int import from_zero, op_addition
from .char import char_code_at
from .decimal_ import from_parts
from .decimal_ import op_addition as op_addition_1
from .long import op_addition as op_addition_2
from .seq import delay, unfold
from .types import int64, uint8, uint64
from .util import IEnumerable_1, compare


_T = TypeVar("_T")


def make_range_step_function(
    step: Any, stop: Any, zero: Any, add: Callable[[_T, _T], _T]
) -> Callable[[_T], tuple[_T, _T] | None]:
    step_compared_with_zero: int = compare(step, zero) or 0
    if step_compared_with_zero == 0:
        raise Exception("The step of a range cannot be zero")

    step_greater_than_zero: bool = step_compared_with_zero > 0

    def _arrow100(
        x: _T | None = None, step: Any = step, stop: Any = stop, zero: Any = zero, add: Any = add
    ) -> tuple[_T, _T] | None:
        compared_with_last: int = compare(x, stop) or 0
        return (
            ((x, add(x, step)))
            if (
                True
                if ((compared_with_last <= 0) if step_greater_than_zero else False)
                else ((compared_with_last >= 0) if (not step_greater_than_zero) else False)
            )
            else None
        )

    return _arrow100


def integral_range_step(start: Any, step: Any, stop: Any, zero: Any, add: Callable[[_T, _T], _T]) -> IEnumerable_1[Any]:
    step_fn: Callable[[_T], tuple[_T, _T] | None] = make_range_step_function(step, stop, zero, add)

    def _arrow101(
        __unit: None = None, start: Any = start, step: Any = step, stop: Any = stop, zero: Any = zero, add: Any = add
    ) -> IEnumerable_1[_T]:
        return unfold(step_fn, start)

    return delay(_arrow101)


def range_big_int(start: int, step: int, stop: int) -> IEnumerable_1[int]:
    def _arrow102(x: int, y: int, start: Any = start, step: Any = step, stop: Any = stop) -> int:
        return op_addition(x, y)

    return integral_range_step(start, step, stop, from_zero(), _arrow102)


def range_decimal(start: Decimal, step: Decimal, stop: Decimal) -> IEnumerable_1[Decimal]:
    def _arrow104(x: Decimal, y: Decimal, start: Any = start, step: Any = step, stop: Any = stop) -> Decimal:
        return op_addition_1(x, y)

    return integral_range_step(start, step, stop, from_parts(0, 0, 0, False, uint8(0)), _arrow104)


def range_double(start: float, step: float, stop: float) -> IEnumerable_1[float]:
    def _arrow105(x: float, y: float, start: Any = start, step: Any = step, stop: Any = stop) -> float:
        return x + y

    return integral_range_step(start, step, stop, 0.0, _arrow105)


def range_int64(start: int64, step: int64, stop: int64) -> IEnumerable_1[int64]:
    def _arrow106(x: int64, y: int64, start: Any = start, step: Any = step, stop: Any = stop) -> int64:
        return op_addition_2(x, y)

    return integral_range_step(start, step, stop, int64(0), _arrow106)


def range_uint64(start: uint64, step: uint64, stop: uint64) -> IEnumerable_1[uint64]:
    def _arrow107(x: uint64, y: uint64, start: Any = start, step: Any = step, stop: Any = stop) -> uint64:
        return op_addition_2(x, y)

    return integral_range_step(start, step, stop, uint64(0), _arrow107)


def range_char(start: str, stop: str) -> IEnumerable_1[str]:
    int_stop: int = int(char_code_at(stop, 0)) or 0

    def _arrow109(__unit: None = None, start: Any = start, stop: Any = stop) -> IEnumerable_1[str]:
        def step_fn(c: int) -> tuple[str, int] | None:
            if c <= int_stop:
                return (chr(c), c + 1)

            else:
                return None

        return unfold(step_fn, int(char_code_at(start, 0)))

    return delay(_arrow109)


__all__ = [
    "make_range_step_function",
    "integral_range_step",
    "range_big_int",
    "range_decimal",
    "range_double",
    "range_int64",
    "range_uint64",
    "range_char",
]



from __future__ import annotations

import time


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class StopWatch:
    def __init__(self) -> None:
        self._start_time: float | None = None
        self._elapsed_time: float | None = None

    def is_running(self) -> bool:
        """Return True if timer is running"""
        return self._start_time is not None

    def start(self) -> None:
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError("Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self) -> None:
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        self._elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None

    def elapsed_milliseconds(self) -> int:
        """Return the elapsed time in milliseconds"""
        if self._elapsed_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        return int(self._elapsed_time * 1000)

    def elapsed_ticks(self) -> int:
        """Return the elapsed time in ticks"""
        if self._elapsed_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        return int(self._elapsed_time * 10000000)


def frequency() -> int:
    return 1000000000


def get_timestamp() -> int:
    return time.perf_counter_ns()



from dataclasses import dataclass
from typing import Any
from ..fable_library.reflection import (TypeInfo, string_type, class_type, record_type)
from ..fable_library.types import Record

def _expr0() -> TypeInfo:
    return record_type("ARCtrl.FileSystem.Commit", [], Commit, lambda: [("Hash", string_type), ("UserName", string_type), ("UserEmail", string_type), ("Date", class_type("System.DateTime")), ("Message", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class Commit(Record):
    Hash: str
    UserName: str
    UserEmail: str
    Date: Any
    Message: str

Commit_reflection = _expr0

__all__ = ["Commit_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.data_context import (DataContext, DataContext__get_Explication, DataContext__get_Unit, DataContext__get_ObjectType, DataContext__get_Description, DataContext__get_GeneratedBy, DataContext__get_Label)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (map, FSharpList, of_array, singleton, transpose)
from ...fable_library.option import (default_arg, bind)
from ...fable_library.seq import (to_list, collect, map as map_1, delay, append, singleton as singleton_1, try_find)
from ...fable_library.seq2 import distinct
from ...fable_library.types import Array
from ...fable_library.util import (ignore, IEnumerable_1, string_hash)
from ...fs_spreadsheet.Cells.fs_cell import FsCell
from ...fs_spreadsheet.fs_column import FsColumn
from .data_map_header import (from_fs_cells, to_fs_cells)

def set_from_fs_columns(dc: Array[DataContext], columns: FSharpList[FsColumn]) -> Array[DataContext]:
    def mapping(c: FsColumn, dc: Any=dc, columns: Any=columns) -> FsCell:
        return c.Item(1)

    cell_parser: Callable[[DataContext, FSharpList[FsCell]], DataContext] = from_fs_cells(map(mapping, columns))
    for i in range(0, (len(dc) - 1) + 1, 1):
        def mapping_1(c_1: FsColumn, dc: Any=dc, columns: Any=columns) -> FsCell:
            return c_1.Item(i + 2)

        ignore(cell_parser(dc[i])(map(mapping_1, columns)))
    return dc


def to_fs_columns(dc: Array[DataContext]) -> FSharpList[FSharpList[FsCell]]:
    def mapping_1(dc_1: DataContext, dc: Any=dc) -> IEnumerable_1[str]:
        def mapping(c: Comment, dc_1: Any=dc_1) -> str:
            return default_arg(c.Name, "")

        return map_1(mapping, dc_1.Comments)

    class ObjectExpr3596:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3594(x: str, y: str) -> bool:
                return x == y

            return _arrow3594

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    comment_keys: FSharpList[str] = to_list(distinct(collect(mapping_1, dc), ObjectExpr3596()))
    headers: FSharpList[FsCell] = to_fs_cells(comment_keys)
    def create_term(oa: OntologyAnnotation | None=None, dc: Any=dc) -> FSharpList[FsCell]:
        if oa is None:
            return of_array([FsCell(""), FsCell(""), FsCell("")])

        else: 
            oa_1: OntologyAnnotation = oa
            return of_array([FsCell(default_arg(oa_1.Name, "")), FsCell(default_arg(oa_1.TermSourceREF, "")), FsCell(default_arg(oa_1.TermAccessionNumber, ""))])


    def create_text(s: str | None=None, dc: Any=dc) -> FSharpList[FsCell]:
        return singleton(FsCell(default_arg(s, "")))

    def _arrow3612(__unit: None=None, dc: Any=dc) -> IEnumerable_1[FSharpList[FsCell]]:
        def _arrow3611(__unit: None=None) -> IEnumerable_1[FSharpList[FsCell]]:
            def _arrow3610(dc_4: DataContext) -> FSharpList[FsCell]:
                dc_3: DataContext = dc_4
                def _arrow3609(__unit: None=None) -> IEnumerable_1[FsCell]:
                    def _arrow3599(__unit: None=None) -> FSharpList[FsCell]:
                        dc_2: DataContext = dc_3
                        return of_array([FsCell(default_arg(dc_2.Name, "")), FsCell(default_arg(dc_2.Format, "")), FsCell(default_arg(dc_2.SelectorFormat, ""))])

                    def _arrow3608(__unit: None=None) -> IEnumerable_1[FsCell]:
                        def _arrow3607(__unit: None=None) -> IEnumerable_1[FsCell]:
                            def _arrow3606(__unit: None=None) -> IEnumerable_1[FsCell]:
                                def _arrow3605(__unit: None=None) -> IEnumerable_1[FsCell]:
                                    def _arrow3604(__unit: None=None) -> IEnumerable_1[FsCell]:
                                        def _arrow3603(__unit: None=None) -> IEnumerable_1[FsCell]:
                                            def _arrow3602(__unit: None=None) -> IEnumerable_1[FsCell]:
                                                def mapping_2(key: str) -> FsCell:
                                                    def binder(c_2: Comment, key: Any=key) -> str | None:
                                                        return c_2.Value

                                                    def predicate(c_1: Comment, key: Any=key) -> bool:
                                                        return default_arg(c_1.Name, "") == key

                                                    return FsCell(default_arg(bind(binder, try_find(predicate, dc_3.Comments)), ""))

                                                return map(mapping_2, comment_keys)

                                            return append(create_text(DataContext__get_Label(dc_3)), delay(_arrow3602))

                                        return append(create_text(DataContext__get_GeneratedBy(dc_3)), delay(_arrow3603))

                                    return append(create_text(DataContext__get_Description(dc_3)), delay(_arrow3604))

                                return append(create_term(DataContext__get_ObjectType(dc_3)), delay(_arrow3605))

                            return append(create_term(DataContext__get_Unit(dc_3)), delay(_arrow3606))

                        return append(create_term(DataContext__get_Explication(dc_3)), delay(_arrow3607))

                    return append(_arrow3599(), delay(_arrow3608))

                return to_list(delay(_arrow3609))

            return map_1(_arrow3610, dc)

        return append(singleton_1(headers), delay(_arrow3611))

    return transpose(to_list(delay(_arrow3612)))


__all__ = ["set_from_fs_columns", "to_fs_columns"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.data import Data
from ...arctrl_core.data_context import (DataContext__set_Explication_279AAFF2, DataContext, DataContext__set_Unit_279AAFF2, DataContext__set_ObjectType_279AAFF2, DataContext__set_Description_6DFDD678, DataContext__set_GeneratedBy_6DFDD678, DataContext__set_Label_6DFDD678)
from ...arctrl_core.Helper.collections_ import Option_fromValueWithDefault
from ...arctrl_core.Helper.regex import ActivePatterns__007CComment_007C__007C
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (map, FSharpList, item, is_empty, head, tail, try_find_index)
from ...fable_library.option import map as map_1
from ...fable_library.seq import (to_list, delay, append, map as map_2)
from ...fable_library.string_ import (starts_with_exact, join, to_fail, printf)
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...fs_spreadsheet.Cells.fs_cell import FsCell

def ActivePattern_ontologyAnnotationFromFsCells(tsr_col: int | None, tan_col: int | None, cells: FSharpList[FsCell]) -> OntologyAnnotation:
    def mapping(c: FsCell, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    def _arrow3593(i: int, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cells: Any=cells) -> str:
        return item(i, cell_values)

    tsr: str | None = map_1(_arrow3593, tsr_col)
    def _arrow3595(i_1: int, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cells: Any=cells) -> str:
        return item(i_1, cell_values)

    tan: str | None = map_1(_arrow3595, tan_col)
    return OntologyAnnotation(item(0, cell_values), tsr, tan)


def ActivePattern_freeTextFromFsCells(cells: FSharpList[FsCell]) -> str:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    return item(0, map(mapping, cells))


def ActivePattern_dataFromFsCells(format: int | None, selector_format: int | None, cells: FSharpList[FsCell]) -> Data:
    def mapping(c: FsCell, format: Any=format, selector_format: Any=selector_format, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    def _arrow3597(i: int, format: Any=format, selector_format: Any=selector_format, cells: Any=cells) -> str:
        return item(i, cell_values)

    format_1: str | None = map_1(_arrow3597, format)
    def _arrow3598(i_1: int, format: Any=format, selector_format: Any=selector_format, cells: Any=cells) -> str:
        return item(i_1, cell_values)

    selector_format_1: str | None = map_1(_arrow3598, selector_format)
    return Data(None, item(0, cell_values), None, format_1, selector_format_1)


def ActivePattern__007CTerm_007C__007C(category_string: str, cells: FSharpList[FsCell]) -> Callable[[FSharpList[FsCell]], OntologyAnnotation] | None:
    def _007CAC_007C__007C(s: str, category_string: Any=category_string, cells: Any=cells) -> int | None:
        if s == category_string:
            return 1

        else: 
            return None


    def _007CTSRColumnHeaderRaw_007C__007C(s_1: str, category_string: Any=category_string, cells: Any=cells) -> str | None:
        if starts_with_exact(s_1, "Term Source REF"):
            return s_1

        else: 
            return None


    def _007CTANColumnHeaderRaw_007C__007C(s_2: str, category_string: Any=category_string, cells: Any=cells) -> str | None:
        if starts_with_exact(s_2, "Term Accession Number"):
            return s_2

        else: 
            return None


    def mapping(c: FsCell, category_string: Any=category_string, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result,) = (None,)
    if not is_empty(cell_values):
        active_pattern_result: int | None = _007CAC_007C__007C(head(cell_values))
        if active_pattern_result is not None:
            if is_empty(tail(cell_values)):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(cells_1: FSharpList[FsCell], category_string: Any=category_string, cells: Any=cells) -> OntologyAnnotation:
            return ActivePattern_ontologyAnnotationFromFsCells(None, None, cells_1)

        return Value

    elif pattern_matching_result == 1:
        (pattern_matching_result_1,) = (None,)
        if not is_empty(cell_values):
            active_pattern_result_1: int | None = _007CAC_007C__007C(head(cell_values))
            if active_pattern_result_1 is not None:
                if not is_empty(tail(cell_values)):
                    if _007CTSRColumnHeaderRaw_007C__007C(head(tail(cell_values))) is not None:
                        if not is_empty(tail(tail(cell_values))):
                            if _007CTANColumnHeaderRaw_007C__007C(head(tail(tail(cell_values)))) is not None:
                                if is_empty(tail(tail(tail(cell_values)))):
                                    pattern_matching_result_1 = 0

                                else: 
                                    pattern_matching_result_1 = 1


                            else: 
                                pattern_matching_result_1 = 1


                        else: 
                            pattern_matching_result_1 = 1


                    else: 
                        pattern_matching_result_1 = 1


                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1


        else: 
            pattern_matching_result_1 = 1

        if pattern_matching_result_1 == 0:
            def Value_1(cells_2: FSharpList[FsCell], category_string: Any=category_string, cells: Any=cells) -> OntologyAnnotation:
                return ActivePattern_ontologyAnnotationFromFsCells(1, 2, cells_2)

            return Value_1

        elif pattern_matching_result_1 == 1:
            (pattern_matching_result_2,) = (None,)
            if not is_empty(cell_values):
                active_pattern_result_4: int | None = _007CAC_007C__007C(head(cell_values))
                if active_pattern_result_4 is not None:
                    if not is_empty(tail(cell_values)):
                        if _007CTANColumnHeaderRaw_007C__007C(head(tail(cell_values))) is not None:
                            if not is_empty(tail(tail(cell_values))):
                                if _007CTSRColumnHeaderRaw_007C__007C(head(tail(tail(cell_values)))) is not None:
                                    if is_empty(tail(tail(tail(cell_values)))):
                                        pattern_matching_result_2 = 0

                                    else: 
                                        pattern_matching_result_2 = 1


                                else: 
                                    pattern_matching_result_2 = 1


                            else: 
                                pattern_matching_result_2 = 1


                        else: 
                            pattern_matching_result_2 = 1


                    else: 
                        pattern_matching_result_2 = 1


                else: 
                    pattern_matching_result_2 = 1


            else: 
                pattern_matching_result_2 = 1

            if pattern_matching_result_2 == 0:
                def Value_2(cells_3: FSharpList[FsCell], category_string: Any=category_string, cells: Any=cells) -> OntologyAnnotation:
                    return ActivePattern_ontologyAnnotationFromFsCells(2, 1, cells_3)

                return Value_2

            elif pattern_matching_result_2 == 1:
                return None





def ActivePattern__007CExplication_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    active_pattern_result: Callable[[FSharpList[FsCell]], OntologyAnnotation] | None = ActivePattern__007CTerm_007C__007C("Explication", cells)
    if active_pattern_result is not None:
        r: Callable[[FSharpList[FsCell]], OntologyAnnotation] = active_pattern_result
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3600(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_Explication_279AAFF2(dc, r(cells_1))
                return dc

            return _arrow3600

        return Value

    else: 
        return None



def ActivePattern__007CUnit_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    active_pattern_result: Callable[[FSharpList[FsCell]], OntologyAnnotation] | None = ActivePattern__007CTerm_007C__007C("Unit", cells)
    if active_pattern_result is not None:
        r: Callable[[FSharpList[FsCell]], OntologyAnnotation] = active_pattern_result
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3601(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_Unit_279AAFF2(dc, r(cells_1))
                return dc

            return _arrow3601

        return Value

    else: 
        return None



def ActivePattern__007CObjectType_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    active_pattern_result: Callable[[FSharpList[FsCell]], OntologyAnnotation] | None = ActivePattern__007CTerm_007C__007C("Object Type", cells)
    if active_pattern_result is not None:
        r: Callable[[FSharpList[FsCell]], OntologyAnnotation] = active_pattern_result
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3613(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_ObjectType_279AAFF2(dc, r(cells_1))
                return dc

            return _arrow3613

        return Value

    else: 
        return None



def ActivePattern__007CDescription_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result,) = (None,)
    if not is_empty(cell_values):
        if head(cell_values) == "Description":
            if is_empty(tail(cell_values)):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3614(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_Description_6DFDD678(dc, Option_fromValueWithDefault("", ActivePattern_freeTextFromFsCells(cells_1)))
                return dc

            return _arrow3614

        return Value

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CGeneratedBy_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result,) = (None,)
    if not is_empty(cell_values):
        if head(cell_values) == "Generated By":
            if is_empty(tail(cell_values)):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3615(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_GeneratedBy_6DFDD678(dc, Option_fromValueWithDefault("", ActivePattern_freeTextFromFsCells(cells_1)))
                return dc

            return _arrow3615

        return Value

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CLabel_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result,) = (None,)
    if not is_empty(cell_values):
        if head(cell_values) == "Label":
            if is_empty(tail(cell_values)):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3616(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                DataContext__set_Label_6DFDD678(dc, Option_fromValueWithDefault("", ActivePattern_freeTextFromFsCells(cells_1)))
                return dc

            return _arrow3616

        return Value

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CData_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result, cols) = (None, None)
    if not is_empty(cell_values):
        if head(cell_values) == "Data":
            pattern_matching_result = 0
            cols = tail(cell_values)

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3617(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                def mapping_1(y: int) -> int:
                    return 1 + y

                def predicate(s: str) -> bool:
                    return starts_with_exact(s, "Data Format")

                def mapping_2(y_1: int) -> int:
                    return 1 + y_1

                def predicate_1(s_1: str) -> bool:
                    return starts_with_exact(s_1, "Data Selector Format")

                d: Data = ActivePattern_dataFromFsCells(map_1(mapping_1, try_find_index(predicate, cols)), map_1(mapping_2, try_find_index(predicate_1, cols)), cells_1)
                dc.FilePath = d.FilePath
                dc.Selector = d.Selector
                dc.Format = d.Format
                dc.SelectorFormat = d.SelectorFormat
                return dc

            return _arrow3617

        return Value

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CComment_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result, key) = (None, None)
    if not is_empty(cell_values):
        active_pattern_result: str | None = ActivePatterns__007CComment_007C__007C(head(cell_values))
        if active_pattern_result is not None:
            if is_empty(tail(cell_values)):
                pattern_matching_result = 0
                key = active_pattern_result

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3618(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                def mapping_1(c_1: FsCell) -> str:
                    return c_1.ValueAsString()

                comment: str = item(0, map(mapping_1, cells_1))
                (dc.Comments.append(Comment.create(key, comment)))
                return dc

            return _arrow3618

        return Value

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CFreetext_007C__007C(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext] | None:
    def mapping(c: FsCell, cells: Any=cells) -> str:
        return c.ValueAsString()

    cell_values: FSharpList[str] = map(mapping, cells)
    (pattern_matching_result, key) = (None, None)
    if not is_empty(cell_values):
        if is_empty(tail(cell_values)):
            pattern_matching_result = 0
            key = head(cell_values)

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def Value(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3619(cells_1: FSharpList[FsCell], dc: Any=dc) -> DataContext:
                def mapping_1(c_1: FsCell) -> str:
                    return c_1.ValueAsString()

                comment: str = item(0, map(mapping_1, cells_1))
                (dc.Comments.append(Comment.create(key, comment)))
                return dc

            return _arrow3619

        return Value

    elif pattern_matching_result == 1:
        return None



def from_fs_cells(cells: FSharpList[FsCell]) -> Callable[[DataContext, FSharpList[FsCell]], DataContext]:
    (pattern_matching_result, r) = (None, None)
    active_pattern_result: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CExplication_007C__007C(cells)
    if active_pattern_result is not None:
        pattern_matching_result = 0
        r = active_pattern_result

    else: 
        active_pattern_result_1: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CUnit_007C__007C(cells)
        if active_pattern_result_1 is not None:
            pattern_matching_result = 0
            r = active_pattern_result_1

        else: 
            active_pattern_result_2: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CObjectType_007C__007C(cells)
            if active_pattern_result_2 is not None:
                pattern_matching_result = 0
                r = active_pattern_result_2

            else: 
                active_pattern_result_3: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CDescription_007C__007C(cells)
                if active_pattern_result_3 is not None:
                    pattern_matching_result = 0
                    r = active_pattern_result_3

                else: 
                    active_pattern_result_4: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CGeneratedBy_007C__007C(cells)
                    if active_pattern_result_4 is not None:
                        pattern_matching_result = 0
                        r = active_pattern_result_4

                    else: 
                        active_pattern_result_5: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CLabel_007C__007C(cells)
                        if active_pattern_result_5 is not None:
                            pattern_matching_result = 0
                            r = active_pattern_result_5

                        else: 
                            active_pattern_result_6: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CData_007C__007C(cells)
                            if active_pattern_result_6 is not None:
                                pattern_matching_result = 0
                                r = active_pattern_result_6

                            else: 
                                active_pattern_result_7: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CComment_007C__007C(cells)
                                if active_pattern_result_7 is not None:
                                    pattern_matching_result = 0
                                    r = active_pattern_result_7

                                else: 
                                    active_pattern_result_8: Callable[[DataContext, FSharpList[FsCell]], DataContext] | None = ActivePattern__007CFreetext_007C__007C(cells)
                                    if active_pattern_result_8 is not None:
                                        pattern_matching_result = 0
                                        r = active_pattern_result_8

                                    else: 
                                        pattern_matching_result = 1









    if pattern_matching_result == 0:
        def _arrow3621(dc: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            def _arrow3620(cells_1: FSharpList[FsCell]) -> DataContext:
                return r(dc)(cells_1)

            return _arrow3620

        return _arrow3621

    elif pattern_matching_result == 1:
        def mapping(c: FsCell, cells: Any=cells) -> str:
            return c.ValueAsString()

        arg: str = join(", ", map(mapping, cells))
        clo_1: Callable[[DataContext, FSharpList[FsCell]], DataContext] = to_fail(printf("Could not parse data map column: %s"))(arg)
        def _arrow3622(arg_1: DataContext, cells: Any=cells) -> Callable[[FSharpList[FsCell]], DataContext]:
            clo_2: Callable[[FSharpList[FsCell]], DataContext] = clo_1(arg_1)
            return clo_2

        return _arrow3622



def to_fs_cells(comment_keys: FSharpList[str]) -> FSharpList[FsCell]:
    def _arrow3631(__unit: None=None, comment_keys: Any=comment_keys) -> IEnumerable_1[FsCell]:
        def _arrow3630(__unit: None=None) -> IEnumerable_1[FsCell]:
            def _arrow3629(__unit: None=None) -> IEnumerable_1[FsCell]:
                def _arrow3628(__unit: None=None) -> IEnumerable_1[FsCell]:
                    def _arrow3627(__unit: None=None) -> IEnumerable_1[FsCell]:
                        def _arrow3626(__unit: None=None) -> IEnumerable_1[FsCell]:
                            def _arrow3625(__unit: None=None) -> IEnumerable_1[FsCell]:
                                def _arrow3624(__unit: None=None) -> IEnumerable_1[FsCell]:
                                    def _arrow3623(ck: str) -> FsCell:
                                        return FsCell(("Comment [" + ck) + "]")

                                    return map_2(_arrow3623, comment_keys)

                                return append(to_enumerable([FsCell("Label")]), delay(_arrow3624))

                            return append(to_enumerable([FsCell("Generated By")]), delay(_arrow3625))

                        return append(to_enumerable([FsCell("Description")]), delay(_arrow3626))

                    return append(to_enumerable([FsCell("Object Type"), FsCell("Term Source REF"), FsCell("Term Accession Number")]), delay(_arrow3627))

                return append(to_enumerable([FsCell("Unit"), FsCell("Term Source REF"), FsCell("Term Accession Number")]), delay(_arrow3628))

            return append(to_enumerable([FsCell("Explication"), FsCell("Term Source REF"), FsCell("Term Accession Number")]), delay(_arrow3629))

        return append(to_enumerable([FsCell("Data"), FsCell("Data Format"), FsCell("Data Selector Format")]), delay(_arrow3630))

    return to_list(delay(_arrow3631))


__all__ = ["ActivePattern_ontologyAnnotationFromFsCells", "ActivePattern_freeTextFromFsCells", "ActivePattern_dataFromFsCells", "ActivePattern__007CTerm_007C__007C", "ActivePattern__007CExplication_007C__007C", "ActivePattern__007CUnit_007C__007C", "ActivePattern__007CObjectType_007C__007C", "ActivePattern__007CDescription_007C__007C", "ActivePattern__007CGeneratedBy_007C__007C", "ActivePattern__007CLabel_007C__007C", "ActivePattern__007CData_007C__007C", "ActivePattern__007CComment_007C__007C", "ActivePattern__007CFreetext_007C__007C", "from_fs_cells", "to_fs_cells"]




from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterable
from typing import (
    Any,
    Protocol,
    TypeVar,
    overload,
)

from .task import from_result, zero
from .util import IDisposable


_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_TD = TypeVar("_TD", bound=IDisposable)
_U = TypeVar("_U")


class Delayed(Protocol[_T_co]):
    def __call__(self, __unit: None | None = None) -> Awaitable[_T_co]: ...


class TaskBuilder:
    def Bind(self, computation: Awaitable[_T], binder: Callable[[_T], Awaitable[_U]]) -> Awaitable[_U]:
        async def bind() -> _U:
            value = await computation
            return await binder(value)

        return bind()

    def Combine(self, computation1: Awaitable[None], computation2: Delayed[_T]) -> Awaitable[_T]:
        return self.Bind(computation1, computation2)

    def Delay(self, generator: Callable[[], Awaitable[_T]]) -> Delayed[_T]:
        def deferred(_: Any = None) -> Awaitable[_T]:
            # print("Delay: deferred: ", generator)
            return generator()

        return deferred

    def For(self, sequence: Iterable[_T], body: Callable[[_T], Awaitable[_U]]) -> Awaitable[_U]:
        done = False
        it = iter(sequence)
        try:
            cur = next(it)
        except StopIteration:
            done = True

        def delay():
            nonlocal cur, done
            res = body(cur)
            try:
                cur = next(it)
            except StopIteration:
                done = True
            return res

        return self.While(lambda: not done, self.Delay(delay))

    @overload
    def Return(self) -> Awaitable[None]: ...

    @overload
    def Return(self, value: _T) -> Awaitable[_T]: ...

    def Return(self, value: Any = None) -> Awaitable[Any]:
        return from_result(value)

    def ReturnFrom(self, computation: Awaitable[_T]) -> Awaitable[_T]:
        return computation

    def TryFinally(self, computation: Delayed[_T], compensation: Callable[[], None]) -> Awaitable[_T]:
        async def try_finally() -> _T:
            try:
                t = await computation()
            finally:
                compensation()
            return t

        return try_finally()

    def TryWith(self, computation: Delayed[_T], catchHandler: Callable[[Any], Awaitable[_T]]) -> Awaitable[_T]:
        async def try_with() -> _T:
            try:
                t = await computation()
            except Exception as exn:
                t = await catchHandler(exn)
            return t

        return try_with()

    def Using(self, resource: _TD, binder: Callable[[_TD], Awaitable[_U]]) -> Awaitable[_U]:
        return self.TryFinally(self.Delay(lambda: binder(resource)), lambda: resource.Dispose())

    @overload
    def While(self, guard: Callable[[], bool], computation: Delayed[None]) -> Awaitable[None]: ...

    @overload
    def While(self, guard: Callable[[], bool], computation: Delayed[_T]) -> Awaitable[_T]: ...

    def While(self, guard: Callable[[], bool], computation: Delayed[Any]) -> Awaitable[Any]:
        if guard():
            return self.Bind(computation(), lambda _: self.While(guard, computation))
        else:
            return self.Return()

    def Zero(self) -> Awaitable[None]:
        return zero()

    def Run(self, computation: Delayed[_T]) -> Awaitable[_T]:
        # Make sure we don't execute computation right now, so wrap in a coroutine.
        async def run() -> _T:
            return await computation()

        return run()


task = TaskBuilder

__all__ = ["task"]



from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1619() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Component.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Component", string_type), ("ArcComponent", string_type), ("component_name", string_type), ("component_type", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Component: str
    ArcComponent: str
    component_name: str
    component_type: str

IContext_reflection = _expr1619

def _arrow1629(__unit: None=None) -> IEncodable:
    class ObjectExpr1620(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1621(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:PropertyValue")

    class ObjectExpr1622(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:name")

    class ObjectExpr1623(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:propertyID")

    class ObjectExpr1624(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:value")

    class ObjectExpr1625(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:valueReference")

    class ObjectExpr1626(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:unitText")

    class ObjectExpr1627(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:unitCode")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1620()), ("Component", ObjectExpr1621()), ("category", ObjectExpr1622()), ("categoryCode", ObjectExpr1623()), ("value", ObjectExpr1624()), ("valueCode", ObjectExpr1625()), ("unit", ObjectExpr1626()), ("unitCode", ObjectExpr1627())])
    class ObjectExpr1628(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_8))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_8.encode_object(arg)

    return ObjectExpr1628()


context_jsonvalue: IEncodable = _arrow1629()

context_str: str = "\r\n{\r\n  \"@context\": {\r\n    \"sdo\": \"http://schema.org/\",\r\n    \r\n    \"Component\": \"sdo:PropertyValue\",\r\n\r\n    \"componentName\": \"sdo\",\r\n    \"componentType\": \"arc:ARC#ARC_00000102\"\r\n  }\r\n}\r\n    "

__all__ = ["IContext_reflection", "context_jsonvalue", "context_str"]




from __future__ import annotations
from typing import Any
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import LDNode

def _expr1091() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDDefinedTerm", None, LDDefinedTerm)


class LDDefinedTerm:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/DefinedTerm"

    @staticmethod
    def term_code() -> str:
        return "http://schema.org/termCode"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def try_get_term_code_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDDefinedTerm.term_code(), context)
        (pattern_matching_result, tc) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                tc = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return tc

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_term_code_as_string(dt: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDDefinedTerm.term_code(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                tc: str = value(match_value)
                return tc

            else: 
                raise Exception(("Property of `termCode` of object with @id `" + dt.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `termCode` of object with @id `" + dt.Id) + "`")


    @staticmethod
    def set_term_code_as_string(dt: LDNode, term_code: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDDefinedTerm.term_code(), term_code, context)

    @staticmethod
    def try_get_name_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDDefinedTerm.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(dt: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDDefinedTerm.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + dt.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + dt.Id) + "`")


    @staticmethod
    def set_name_as_string(dt: LDNode, name: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDDefinedTerm.name(), name, context)

    @staticmethod
    def gen_id(name: str, term_code: str | None=None) -> str:
        return clean(("#OA_" + name) + "") if (term_code is None) else (("" + term_code) + "")

    @staticmethod
    def validate(dt: LDNode, context: LDContext | None=None) -> bool:
        return dt.HasProperty(LDDefinedTerm.name(), context) if dt.HasType(LDDefinedTerm.schema_type(), context) else False

    @staticmethod
    def create(name: str, id: str | None=None, term_code: str | None=None, context: LDContext | None=None) -> LDNode:
        dt: LDNode = LDNode(LDDefinedTerm.gen_id(name, term_code) if (id is None) else id, [LDDefinedTerm.schema_type()], None, context)
        dt.SetProperty(LDDefinedTerm.name(), name, context)
        dt.SetOptionalProperty(LDDefinedTerm.term_code(), term_code, context)
        return dt


LDDefinedTerm_reflection = _expr1091

__all__ = ["LDDefinedTerm_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_core.Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq)
from ..fable_library.array_ import sort_by
from ..fable_library.option import value
from ..fable_library.range import range_big_int
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (to_list, delay, append, singleton, empty, map, length, fold, item)
from ..fable_library.string_ import join
from ..fable_library.types import (Array, to_string)
from ..fable_library.util import (IEnumerable_1, compare_arrays, equals)
from .validation_package import ValidationPackage

def _expr1004() -> TypeInfo:
    return class_type("ARCtrl.ValidationPackages.ValidationPackagesConfig", None, ValidationPackagesConfig)


class ValidationPackagesConfig:
    def __init__(self, validation_packages: Array[ValidationPackage], arc_specification: str | None=None) -> None:
        self._arc_specification: str | None = arc_specification
        self._validation_packages: Array[ValidationPackage] = validation_packages

    @property
    def ValidationPackages(self, __unit: None=None) -> Array[ValidationPackage]:
        this: ValidationPackagesConfig = self
        return this._validation_packages

    @ValidationPackages.setter
    def ValidationPackages(self, validation_packages: Array[ValidationPackage]) -> None:
        this: ValidationPackagesConfig = self
        this._validation_packages = validation_packages

    @property
    def ARCSpecification(self, __unit: None=None) -> str | None:
        this: ValidationPackagesConfig = self
        return this._arc_specification

    @ARCSpecification.setter
    def ARCSpecification(self, arc_specification: str | None=None) -> None:
        this: ValidationPackagesConfig = self
        this._arc_specification = arc_specification

    @staticmethod
    def make(validation_packages: Array[ValidationPackage], arc_specification: str | None=None) -> ValidationPackagesConfig:
        return ValidationPackagesConfig(validation_packages, arc_specification)

    def Copy(self, __unit: None=None) -> ValidationPackagesConfig:
        this: ValidationPackagesConfig = self
        validation_packages: Array[ValidationPackage] = this.ValidationPackages
        arc_specification: str | None = this.ARCSpecification
        return ValidationPackagesConfig.make(validation_packages, arc_specification)

    def __str__(self, __unit: None=None) -> str:
        this: ValidationPackagesConfig = self
        def _arrow1000(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow999(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow998(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow997(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow996(__unit: None=None) -> IEnumerable_1[str]:
                            def _arrow995(__unit: None=None) -> IEnumerable_1[str]:
                                return singleton("}")

                            return append(singleton("]"), delay(_arrow995))

                        return append(singleton(join((";" + "\n") + "", map(to_string, this.ValidationPackages))), delay(_arrow996))

                    return append(singleton(" ValidationPackages = ["), delay(_arrow997))

                return append(singleton((" ARCSpecification = " + value(this.ARCSpecification)) + "") if (this.ARCSpecification is not None) else empty(), delay(_arrow998))

            return append(singleton("{"), delay(_arrow999))

        return join("\n", to_list(delay(_arrow1000)))

    def StructurallyEquals(self, other: ValidationPackagesConfig) -> bool:
        this: ValidationPackagesConfig = self
        def sort(arg: Array[ValidationPackage]) -> Array[ValidationPackage]:
            def projection(vp: ValidationPackage, arg: Any=arg) -> tuple[str, str | None]:
                return (vp.Name, vp.Version)

            class ObjectExpr1001:
                @property
                def Compare(self) -> Callable[[tuple[str, str | None], tuple[str, str | None]], int]:
                    return compare_arrays

            return sort_by(projection, list(arg), ObjectExpr1001())

        if equals(this.ARCSpecification, other.ARCSpecification):
            a: IEnumerable_1[ValidationPackage] = sort(this.ValidationPackages)
            b: IEnumerable_1[ValidationPackage] = sort(other.ValidationPackages)
            def folder(acc: bool, e: bool) -> bool:
                if acc:
                    return e

                else: 
                    return False


            def _arrow1003(__unit: None=None) -> IEnumerable_1[bool]:
                def _arrow1002(i: int) -> bool:
                    return equals(item(i, a), item(i, b))

                return map(_arrow1002, range_big_int(0, 1, length(a) - 1))

            return fold(folder, True, to_list(delay(_arrow1003))) if (length(a) == length(b)) else False

        else: 
            return False


    def ReferenceEquals(self, other: ValidationPackagesConfig) -> bool:
        this: ValidationPackagesConfig = self
        return this is other

    def __eq__(self, other: Any=None) -> bool:
        this: ValidationPackagesConfig = self
        return this.StructurallyEquals(other) if isinstance(other, ValidationPackagesConfig) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: ValidationPackagesConfig = self
        return box_hash_array([box_hash_option(this.ARCSpecification), box_hash_seq(this.ValidationPackages)])


ValidationPackagesConfig_reflection = _expr1004

def ValidationPackagesConfig__ctor_376974AD(validation_packages: Array[ValidationPackage], arc_specification: str | None=None) -> ValidationPackagesConfig:
    return ValidationPackagesConfig(validation_packages, arc_specification)


__all__ = ["ValidationPackagesConfig_reflection"]

